# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    List, Optional, Union,
)

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class DashboardESDocument(ElasticsearchDocument):
    """
    Schema for the ES dashboard ES document
    """

    def __init__(self,
                 group_name: str,
                 name: str,
                 description: Union[str, None],
                 total_usage: int,
                 product: Optional[str] = '',
                 cluster: Optional[str] = '',
                 group_description: Optional[str] = None,
                 query_names: Union[List[str], None] = None,
                 chart_names: Optional[List[str]] = None,
                 group_url: Optional[str] = None,
                 url: Optional[str] = None,
                 uri: Optional[str] = None,
                 last_successful_run_timestamp: Optional[int] = None,
                 tags: Optional[List[str]] = None,
                 badges: Optional[List[str]] = None,
                 ) -> None:
        self.group_name = group_name
        self.name = name
        self.description = description
        self.cluster = cluster
        self.product = product
        self.group_url = group_url
        self.url = url
        self.uri = uri
        self.last_successful_run_timestamp = last_successful_run_timestamp
        self.total_usage = total_usage
        self.group_description = group_description
        self.query_names = query_names
        self.chart_names = chart_names
        self.tags = tags
        self.badges = badges



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardUsage as RDSDashboardUsage

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.usage.usage_constants import (
    READ_RELATION_COUNT_PROPERTY, READ_RELATION_TYPE, READ_REVERSE_RELATION_TYPE,
)
from databuilder.models.user import User

LOGGER = logging.getLogger(__name__)


class DashboardUsage(GraphSerializable, TableSerializable):
    """
    A model that encapsulate Dashboard usage between Dashboard and User
    """

    def __init__(self,
                 dashboard_group_id: Optional[str],
                 dashboard_id: Optional[str],
                 email: str,
                 view_count: int,
                 should_create_user_node: Optional[bool] = False,
                 product: Optional[str] = '',
                 cluster: Optional[str] = 'gold',
                 **kwargs: Any
                 ) -> None:
        """

        :param dashboard_group_id:
        :param dashboard_id:
        :param email:
        :param view_count:
        :param should_create_user_node: Enable this if it is fine to create/update User node with only with email
        address. Please be advised that other fields will be emptied. Current use case is to create anonymous user.
        For example, Mode dashboard does not provide which user viewed the dashboard and anonymous user can be used
        to show the usage.
        :param product:
        :param cluster:
        :param kwargs:
        """
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        self._email = email
        self._view_count = int(view_count)
        self._product = product
        self._cluster = cluster
        self._user_model = User(email=email)
        self._should_create_user_node = bool(should_create_user_node)
        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        if self._should_create_user_node:
            return self._user_model.create_next_node()

        return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relationship = GraphRelationship(
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            end_label=User.USER_NODE_LABEL,
            start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            ),
            end_key=User.get_user_model_key(email=self._email),
            type=READ_REVERSE_RELATION_TYPE,
            reverse_type=READ_RELATION_TYPE,
            attributes={
                READ_RELATION_COUNT_PROPERTY: self._view_count
            }
        )
        yield relationship

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        if self._should_create_user_node:
            user_record = self._user_model.create_next_record()
            if user_record:
                yield user_record

        dashboard_usage_record = RDSDashboardUsage(
            user_rk=User.get_user_model_key(email=self._email),
            dashboard_rk=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group=self._dashboard_group_id,
                dashboard_name=self._dashboard_id
            ),
            read_count=self._view_count
        )
        yield dashboard_usage_record

    def __repr__(self) -> str:
        return f'DashboardUsage({self._dashboard_group_id!r}, {self._dashboard_id!r}, ' \
               f'{self._email!r}, {self._view_count!r}, {self._should_create_user_node!r}, ' \
               f'{self._product!r}, {self._cluster!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Dict, Iterator,
)

from pyhocon import ConfigFactory, ConfigTree

import databuilder.extractor.dashboard.tableau.tableau_dashboard_constants as const
from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.tableau.tableau_dashboard_utils import (
    TableauDashboardUtils, TableauGraphQLApiExtractor,
)
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT
from databuilder.models.table_metadata import TableMetadata
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel

LOGGER = logging.getLogger(__name__)


class TableauGraphQLDashboardTableExtractor(TableauGraphQLApiExtractor):
    """
    Implements the extraction-time logic for parsing the GraphQL result and transforming into a dict
    that fills the DashboardTable model. Allows workbooks to be exlcuded based on their project.
    """

    CLUSTER = const.CLUSTER
    DATABASE = const.DATABASE
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    EXTERNAL_CLUSTER_NAME = const.EXTERNAL_CLUSTER_NAME

    def execute(self) -> Iterator[Dict[str, Any]]:
        response = self.execute_query()

        workbooks_data = [workbook for workbook in response['workbooks']
                          if workbook['projectName'] not in
                          self._conf.get_list(TableauGraphQLDashboardTableExtractor.EXCLUDED_PROJECTS, [])]

        for workbook in workbooks_data:
            data = {
                'dashboard_group_id': workbook['projectName'],
                'dashboard_id': TableauDashboardUtils.sanitize_workbook_name(workbook['name']),
                'cluster': self._conf.get_string(TableauGraphQLDashboardTableExtractor.CLUSTER),
                'table_ids': []
            }

            for table in workbook['upstreamTables']:

                # external tables have no schema, so they must be parsed differently
                # see TableauExternalTableExtractor for more specifics
                if table['schema'] != '':
                    cluster = self._conf.get_string(TableauGraphQLDashboardTableExtractor.CLUSTER)
                    database = self._conf.get_string(TableauGraphQLDashboardTableExtractor.DATABASE)

                    # Tableau sometimes incorrectly assigns the "schema" value
                    # based on how the datasource connection is used in a workbook.
                    # It will hide the real schema inside the table name, like "real_schema.real_table",
                    # and set the "schema" value to "wrong_schema". In every case discovered so far, the schema
                    # key is incorrect, so the "inner" schema from the table name is used instead.
                    if '.' in table['name']:
                        schema, name = table['name'].split('.')
                    else:
                        schema, name = table['schema'], table['name']
                    schema = TableauDashboardUtils.sanitize_schema_name(schema)
                    name = TableauDashboardUtils.sanitize_table_name(name)
                else:
                    cluster = self._conf.get_string(TableauGraphQLDashboardTableExtractor.EXTERNAL_CLUSTER_NAME)
                    database = TableauDashboardUtils.sanitize_database_name(
                        table['database']['connectionType']
                    )
                    schema = TableauDashboardUtils.sanitize_schema_name(table['database']['name'])
                    name = TableauDashboardUtils.sanitize_table_name(table['name'])

                table_id = TableMetadata.TABLE_KEY_FORMAT.format(
                    db=database,
                    cluster=cluster,
                    schema=schema,
                    tbl=name,
                )
                data['table_ids'].append(table_id)

            yield data


class TableauDashboardTableExtractor(Extractor):
    """
    Extracts metadata about the tables associated with Tableau workbooks.
    It can handle both "regular" database tables as well as "external" tables
    (see TableauExternalTableExtractor for more info on external tables).
    Assumes that all the nodes for both the dashboards and the tables have already been created.
    """

    API_BASE_URL = const.API_BASE_URL
    API_VERSION = const.API_VERSION
    CLUSTER = const.CLUSTER
    DATABASE = const.DATABASE
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    EXTERNAL_CLUSTER_NAME = const.EXTERNAL_CLUSTER_NAME
    SITE_NAME = const.SITE_NAME
    TABLEAU_ACCESS_TOKEN_NAME = const.TABLEAU_ACCESS_TOKEN_NAME
    TABLEAU_ACCESS_TOKEN_SECRET = const.TABLEAU_ACCESS_TOKEN_SECRET
    VERIFY_REQUEST = const.VERIFY_REQUEST

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self.query = """query {
          workbooks {
            name
            projectName
            upstreamTables {
              name
              schema
              database {
                name
                connectionType
              }
            }
          }
        }"""
        self._extractor = self._build_extractor()

        transformers = []
        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_table.DashboardTable'})))
        transformers.append(dict_to_model_transformer)
        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()
        if not record:
            return None

        return next(self._transformer.transform(record=record), None)

    def get_scope(self) -> str:
        return 'extractor.tableau_dashboard_table'

    def _build_extractor(self) -> TableauGraphQLDashboardTableExtractor:
        """
        Builds a TableauGraphQLDashboardTableExtractor. All data required can be retrieved with a single GraphQL call.
        :return: A TableauGraphQLDashboardTableExtractor that creates dashboard <> table relationships.
        """
        extractor = TableauGraphQLDashboardTableExtractor()
        tableau_extractor_conf = \
            Scoped.get_scoped_conf(self._conf, extractor.get_scope())\
                  .with_fallback(self._conf)\
                  .with_fallback(ConfigFactory.from_dict({TableauGraphQLApiExtractor.QUERY: self.query,
                                                          STATIC_RECORD_DICT: {'product': 'tableau'}
                                                          }
                                                         )
                                 )
        extractor.init(conf=tableau_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging
from typing import (
    Any, Callable, Dict, Iterator, List, Optional,
)

from amundsen_rds.models.badge import Badge
from amundsen_rds.models.cluster import Cluster
from amundsen_rds.models.column import ColumnDescription, TableColumn
from amundsen_rds.models.dashboard import (
    Dashboard, DashboardChart, DashboardCluster, DashboardDescription, DashboardExecution, DashboardFollower,
    DashboardGroup, DashboardGroupDescription, DashboardOwner, DashboardQuery, DashboardUsage,
)
from amundsen_rds.models.database import Database
from amundsen_rds.models.schema import Schema, SchemaDescription
from amundsen_rds.models.table import (
    Table, TableDescription, TableFollower, TableOwner, TableProgrammaticDescription, TableTimestamp, TableUsage,
)
from amundsen_rds.models.tag import Tag
from amundsen_rds.models.user import User
from pyhocon import ConfigTree
from sqlalchemy import create_engine, func
from sqlalchemy.orm import (
    Session, load_only, sessionmaker, subqueryload,
)

from databuilder.extractor.base_extractor import Extractor

LOGGER = logging.getLogger(__name__)


def _table_search_query(session: Session, table_filter: List, offset: int, limit: int) -> List:
    """
    Table query
    :param session:
    :param table_filter:
    :param offset:
    :param limit:
    :return:
    """
    # table
    query = session.query(Table).filter(*table_filter).options(
        load_only(Table.rk, Table.name, Table.schema_rk)
    )

    # description
    query = query.options(
        subqueryload(Table.description).options(
            load_only(TableDescription.description)
        )
    ).options(
        subqueryload(Table.programmatic_descriptions).options(
            load_only(TableProgrammaticDescription.description)
        )
    )

    # schema, cluster, database
    query = query.options(
        subqueryload(Table.schema).options(
            load_only(Schema.name, Schema.cluster_rk),
            subqueryload(Schema.description).options(
                load_only(SchemaDescription.description)
            ),
            subqueryload(Schema.cluster).options(
                load_only(Cluster.name, Cluster.database_rk),
                subqueryload(Cluster.database).options(
                    load_only(Database.name)
                )
            )
        )
    )

    # column
    query = query.options(
        subqueryload(Table.columns).options(
            load_only(TableColumn.rk, TableColumn.name),
            subqueryload(TableColumn.description).options(
                load_only(ColumnDescription.description)
            )
        )
    )

    # tag, badge
    query = query.options(
        subqueryload(Table.tags).options(
            load_only(Tag.rk, Tag.tag_type)
        )
    ).options(
        subqueryload(Table.badges).options(
            load_only(Badge.rk)
        )
    )

    # usage
    query = query.options(
        subqueryload(Table.usage).options(
            load_only(TableUsage.read_count)
        )
    )

    # timestamp
    query = query.options(
        subqueryload(Table.timestamp).options(
            load_only(TableTimestamp.last_updated_timestamp)
        )
    )

    query = query.order_by(Table.rk).offset(offset).limit(limit)

    return query.all()


def _table_search(session: Session, published_tag: str, limit: int) -> List[Dict]:
    """
    Query table metadata.
    :param session:
    :param published_tag:
    :param limit:
    :return:
    """
    LOGGER.info('Querying table metadata.')

    table_filter = []
    if published_tag:
        table_filter.append(Table.published_tag == published_tag)

    table_results = []

    offset = 0
    tables = _table_search_query(session, table_filter, offset, limit)
    while tables:
        for table in tables:
            schema = table.schema
            schema_description = schema.description.description if schema.description else None
            cluster = schema.cluster
            database = cluster.database
            description = table.description.description if table.description else ''
            programmatic_descriptions = [description.description
                                         for description in table.programmatic_descriptions]

            columns = table.columns
            column_names = [column.name for column in columns]
            column_descriptions = [column.description.description if column.description else ''
                                   for column in columns]

            total_usage = sum(usage.read_count for usage in table.usage)
            unique_usage = len(table.usage)

            tags = [tag.rk for tag in table.tags if tag.tag_type == 'default']
            badges = [badge.rk for badge in table.badges]
            last_updated_timestamp = table.timestamp.last_updated_timestamp if table.timestamp else None

            table_result = dict(database=database.name,
                                cluster=cluster.name,
                                schema=schema.name,
                                name=table.name,
                                key=table.rk,
                                description=description,
                                last_updated_timestamp=last_updated_timestamp,
                                column_names=column_names,
                                column_descriptions=column_descriptions,
                                total_usage=total_usage,
                                unique_usage=unique_usage,
                                tags=tags,
                                badges=badges,
                                schema_description=schema_description,
                                programmatic_descriptions=programmatic_descriptions)
            table_results.append(table_result)

        offset += limit
        tables = _table_search_query(session, table_filter, offset, limit)

    return table_results


def _dashboard_search_query(session: Session, dashboard_filter: List, offset: int, limit: int) -> List:
    """
    Dashboard query
    :param session:
    :param dashboard_filter:
    :param offset:
    :param limit:
    :return:
    """
    # dashboard
    query = session.query(Dashboard).filter(*dashboard_filter).options(
        load_only(Dashboard.rk,
                  Dashboard.name,
                  Dashboard.dashboard_url,
                  Dashboard.dashboard_group_rk)
    )

    # group, cluster
    query = query.options(
        subqueryload(Dashboard.group).options(
            load_only(DashboardGroup.rk,
                      DashboardGroup.name,
                      DashboardGroup.dashboard_group_url,
                      DashboardGroup.cluster_rk),
            subqueryload(DashboardGroup.description).options(
                load_only(DashboardGroupDescription.description)
            ),
            subqueryload(DashboardGroup.cluster).options(
                load_only(DashboardCluster.name)
            )
        )
    )

    # description
    query = query.options(
        subqueryload(Dashboard.description).options(
            load_only(DashboardDescription.description)
        )
    )

    # execution
    query = query.options(
        subqueryload(Dashboard.execution).options(
            load_only(DashboardExecution.rk, DashboardExecution.timestamp)
        )
    )

    # usage
    query = query.options(
        subqueryload(Dashboard.usage).options(
            load_only(DashboardUsage.read_count)
        )
    )

    # query, chart
    query = query.options(
        subqueryload(Dashboard.queries).options(
            load_only(DashboardQuery.name),
            subqueryload(DashboardQuery.charts).options(
                load_only(DashboardChart.name)
            )
        )
    )

    # tag, badge
    query = query.options(
        subqueryload(Dashboard.tags).options(
            load_only(Tag.rk, Tag.tag_type)
        )
    ).options(
        subqueryload(Dashboard.badges).options(
            load_only(Badge.rk)
        )
    )

    query = query.order_by(Dashboard.rk).offset(offset).limit(limit)

    return query.all()


def _dashboard_search(session: Session, published_tag: str, limit: int) -> List[Dict]:
    """
    Query dashboard metadata.
    :param session:
    :param published_tag:
    :param limit:
    :return:
    """
    LOGGER.info('Querying dashboard metadata.')

    dashboard_filter = []
    if published_tag:
        dashboard_filter.append(Dashboard.published_tag == published_tag)

    dashboard_results = []

    offset = 0
    dashboards = _dashboard_search_query(session, dashboard_filter, offset, limit)
    while dashboards:
        for dashboard in dashboards:
            group = dashboard.group
            description = dashboard.description.description if dashboard.description else None
            group_description = group.description.description if group.description else None
            cluster = group.cluster
            product = dashboard.rk.split('_')[0]
            last_exec = next((execution for execution in dashboard.execution
                              if execution.rk.endswith('_last_successful_execution')), None)
            last_successful_run_timestamp = last_exec.timestamp if last_exec else None
            total_usage = sum(usage.read_count for usage in dashboard.usage)

            queries = dashboard.queries
            query_names = [query.name for query in queries]
            chart_names = [chart.name for query in queries for chart in query.charts]

            tags = [tag.rk for tag in dashboard.tags if tag.tag_type == 'default']
            badges = [badge.rk for badge in dashboard.badges]

            dashboard_result = dict(group_name=group.name,
                                    name=dashboard.name,
                                    description=description,
                                    total_usage=total_usage,
                                    product=product,
                                    cluster=cluster.name,
                                    group_description=group_description,
                                    query_names=query_names,
                                    chart_names=chart_names,
                                    group_url=group.dashboard_group_url,
                                    url=dashboard.dashboard_url,
                                    uri=dashboard.rk,
                                    last_successful_run_timestamp=last_successful_run_timestamp,
                                    tags=tags,
                                    badges=badges)

            dashboard_results.append(dashboard_result)

        offset += limit
        dashboards = _dashboard_search_query(session, dashboard_filter, offset, limit)

    return dashboard_results


def _user_search_query(session: Session, user_filter: List, offset: int, limit: int) -> List:
    """
    User query
    :param session:
    :param user_filter:
    :param offset:
    :param limit:
    :return:
    """
    # read
    table_usage_subquery = session \
        .query(User.rk, func.sum(TableUsage.read_count).label('table_read_count')) \
        .outerjoin(TableUsage) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    dashboard_usage_subquery = session \
        .query(User.rk, func.sum(DashboardUsage.read_count).label('dashboard_read_count')) \
        .outerjoin(DashboardUsage) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    # own
    table_owner_subquery = session \
        .query(User.rk, func.count(TableOwner.table_rk).label('table_own_count')) \
        .outerjoin(TableOwner) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    dashboard_owner_subquery = session \
        .query(User.rk, func.count(DashboardOwner.dashboard_rk).label('dashboard_own_count')) \
        .outerjoin(DashboardOwner) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    # follow
    table_follower_subquery = session \
        .query(User.rk, func.count(TableFollower.table_rk).label('table_follow_count')) \
        .outerjoin(TableFollower) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    dashboard_follower_subquery = session \
        .query(User.rk, func.count(DashboardFollower.dashboard_rk).label('dashboard_follow_count')) \
        .outerjoin(DashboardFollower) \
        .filter(*user_filter) \
        .group_by(User.rk).order_by(User.rk).offset(offset).limit(limit).subquery()

    # user
    query = session \
        .query(User,
               table_usage_subquery.c.table_read_count,
               dashboard_usage_subquery.c.dashboard_read_count,
               table_owner_subquery.c.table_own_count,
               dashboard_owner_subquery.c.dashboard_own_count,
               table_follower_subquery.c.table_follow_count,
               dashboard_follower_subquery.c.dashboard_follow_count) \
        .join(table_usage_subquery, table_usage_subquery.c.rk == User.rk) \
        .join(dashboard_usage_subquery, dashboard_usage_subquery.c.rk == User.rk) \
        .join(table_owner_subquery, table_owner_subquery.c.rk == User.rk) \
        .join(dashboard_owner_subquery, dashboard_owner_subquery.c.rk == User.rk) \
        .join(table_follower_subquery, table_follower_subquery.c.rk == User.rk) \
        .join(dashboard_follower_subquery, dashboard_follower_subquery.c.rk == User.rk)

    # manager
    query = query.options(
        subqueryload(User.manager).options(
            load_only(User.email)
        )
    )

    return query.all()


def _user_search(session: Session, published_tag: str, limit: int) -> List[Dict]:
    """
    Query user metadata.
    :param session:
    :param published_tag:
    :param limit:
    :return:
    """
    LOGGER.info('Querying user metadata.')

    user_filter = [User.full_name.isnot(None)]
    if published_tag:
        user_filter.append(User.published_tag == published_tag)

    user_results = []

    offset = 0
    query_results = _user_search_query(session, user_filter, offset, limit)
    while query_results:
        for query_result in query_results:
            user = query_result.User
            table_read_count = int(query_result.table_read_count) if query_result.table_read_count else 0
            dashboard_read_count = int(query_result.dashboard_read_count) if query_result.dashboard_read_count else 0
            total_read_count = table_read_count + dashboard_read_count

            table_own_count = query_result.table_own_count
            dashboard_own_count = query_result.dashboard_own_count
            total_own_count = table_own_count + dashboard_own_count

            table_follow_count = query_result.table_follow_count
            dashboard_follow_count = query_result.dashboard_follow_count
            total_follow_count = table_follow_count + dashboard_follow_count

            manager_email = user.manager.email if user.manager else ''
            user_result = dict(email=user.email,
                               first_name=user.first_name,
                               last_name=user.last_name,
                               full_name=user.full_name,
                               github_username=user.github_username,
                               team_name=user.team_name,
                               employee_type=user.employee_type,
                               manager_email=manager_email,
                               slack_id=user.slack_id,
                               role_name=user.role_name,
                               is_active=user.is_active,
                               total_read=total_read_count,
                               total_own=total_own_count,
                               total_follow=total_follow_count)

            user_results.append(user_result)

        offset += limit
        query_results = _user_search_query(session, user_filter, offset, limit)

    return user_results


class MySQLSearchDataExtractor(Extractor):
    """
    Extractor to fetch data required to support search from MySQL.
    """
    ENTITY_TYPE = 'entity_type'
    MODEL_CLASS = 'model_class'
    JOB_PUBLISH_TAG = 'job_publish_tag'
    SEARCH_FUNCTION = 'search_function'

    CONN_STRING = 'conn_string'
    ENGINE_ECHO = 'engine_echo'
    CONNECT_ARGS = 'connect_args'
    QUERY_LIMIT = 'query_limit'

    _DEFAULT_QUERY_LIMIT = 500
    _DEFAULT_SEARCH_BY_ENTITY: Dict[str, Callable] = {
        'table': _table_search,
        'user': _user_search,
        'dashboard': _dashboard_search
    }

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf
        self.entity = conf.get_string(MySQLSearchDataExtractor.ENTITY_TYPE, default='table').lower()
        if MySQLSearchDataExtractor.SEARCH_FUNCTION in conf:
            self.search_function = conf.get(MySQLSearchDataExtractor.SEARCH_FUNCTION)
        else:
            self.search_function = MySQLSearchDataExtractor._DEFAULT_SEARCH_BY_ENTITY[self.entity]
        self.published_tag = conf.get_string(MySQLSearchDataExtractor.JOB_PUBLISH_TAG, '')
        self.query_limit = conf.get_int(MySQLSearchDataExtractor.QUERY_LIMIT, self._DEFAULT_QUERY_LIMIT)

        connect_args = {k: v for k, v in self.conf.get_config(MySQLSearchDataExtractor.CONNECT_ARGS,
                                                              default=ConfigTree()).items()}
        self._engine = create_engine(conf.get_string(MySQLSearchDataExtractor.CONN_STRING),
                                     echo=conf.get_bool(MySQLSearchDataExtractor.ENGINE_ECHO, False),
                                     connect_args=connect_args)
        self._session_factory = sessionmaker(bind=self._engine)

        model_class = conf.get(MySQLSearchDataExtractor.MODEL_CLASS, None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)

        self._extract_iter: Optional[Iterator] = None

    def close(self) -> None:
        """
        Close connection to mysql.
        """
        try:
            self._engine.dispose()
        except Exception as e:
            LOGGER.error(f'Exception encountered while closing engine: {e}')

    def extract(self) -> Optional[Any]:
        """
        Return an object or a raw query result.
        """
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()

        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Iterator[Any]:
        if not hasattr(self, 'results'):
            session = self._session_factory()
            try:
                self.results = self.search_function(session=session,
                                                    published_tag=self.published_tag,
                                                    limit=self.query_limit)
            except Exception as e:
                LOGGER.exception('Exception encountered while executing the search function.')
                raise e
            finally:
                session.close()

        for result in self.results:
            if hasattr(self, 'model_class'):
                obj = self.model_class(**result)
                yield obj
            else:
                yield result

    def get_scope(self) -> str:
        return 'extractor.mysql_search_data'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

from pyhocon import ConfigFactory, ConfigTree
from requests.auth import HTTPBasicAuth

from databuilder import Scoped
from databuilder.extractor.dashboard.mode_analytics.mode_dashboard_constants import (
    MODE_ACCESS_TOKEN, MODE_BEARER_TOKEN, MODE_PASSWORD_TOKEN, ORGANIZATION,
)
from databuilder.extractor.restapi.rest_api_extractor import (
    REST_API_QUERY, STATIC_RECORD_DICT, RestAPIExtractor,
)
from databuilder.rest_api.base_rest_api_query import BaseRestApiQuery, RestApiQuerySeed
from databuilder.rest_api.mode_analytics.mode_paginated_rest_api_query import ModePaginatedRestApiQuery


class ModeDashboardUtils(object):

    @staticmethod
    def get_seed_query(conf: ConfigTree) -> BaseRestApiQuery:
        # Seed query record for next query api to join with
        seed_record = [{'organization': conf.get_string(ORGANIZATION)}]
        seed_query = RestApiQuerySeed(seed_record=seed_record)
        return seed_query

    @staticmethod
    def get_spaces_query_api(conf: ConfigTree) -> BaseRestApiQuery:
        """
        Provides RestApiQuerySeed where it will provides iterator of dictionaries as records where dictionary keys are
         organization, dashboard_group_id, dashboard_group and dashboard_group_description
        :param conf:
        :return:
        """

        # https://mode.com/developer/discovery-api/analytics/spaces
        spaces_url_template = 'https://app.mode.com/batch/{organization}/spaces'

        # Seed query record for next query api to join with
        seed_query = ModeDashboardUtils.get_seed_query(conf=conf)

        # mode_bearer_token must be provided in the conf
        # the token is required to access discovery endpoint
        # https://mode.com/developer/discovery-api/introduction/
        params = ModeDashboardUtils.get_auth_params(conf=conf, discover_auth=True)

        json_path = 'spaces[*].[token,name,description]'
        field_names = ['dashboard_group_id', 'dashboard_group', 'dashboard_group_description']

        # based on https://mode.com/developer/discovery-api/analytics/spaces/#listSpacesForAccount
        pagination_json_path = 'spaces[*]'
        max_per_page = 1000
        spaces_query = ModePaginatedRestApiQuery(pagination_json_path=pagination_json_path,
                                                 max_record_size=max_per_page, query_to_join=seed_query,
                                                 url=spaces_url_template, params=params, json_path=json_path,
                                                 field_names=field_names)

        return spaces_query

    @staticmethod
    def get_auth_params(conf: ConfigTree, discover_auth: bool = False) -> Dict[str, Any]:
        if discover_auth:
            # Mode discovery API needs custom token set in header
            # https://mode.com/developer/discovery-api/introduction/
            params = {
                "headers": {
                    "Authorization": conf.get_string(MODE_BEARER_TOKEN),
                }
            }  # type: Dict[str, Any]
        else:
            params = {
                'auth': HTTPBasicAuth(conf.get_string(MODE_ACCESS_TOKEN),
                                      conf.get_string(MODE_PASSWORD_TOKEN)
                                      )
            }
        return params

    @staticmethod
    def create_mode_rest_api_extractor(restapi_query: BaseRestApiQuery,
                                       conf: ConfigTree
                                       ) -> RestAPIExtractor:
        """
        Creates RestAPIExtractor. Note that RestAPIExtractor is already initialized
        :param restapi_query:
        :param conf:
        :return: RestAPIExtractor. Note that RestAPIExtractor is already initialized
        """
        extractor = RestAPIExtractor()
        rest_api_extractor_conf = \
            Scoped.get_scoped_conf(conf, extractor.get_scope())\
                  .with_fallback(conf)\
                  .with_fallback(ConfigFactory.from_dict({REST_API_QUERY: restapi_query,
                                                          STATIC_RECORD_DICT: {'product': 'mode'}
                                                          }
                                                         )
                                 )

        extractor.init(conf=rest_api_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, List, Optional, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardOwner as RDSDashboardOwner
from amundsen_rds.models.table import TableOwner as RDSTableOwner
from amundsen_rds.models.user import User as RDSUser

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.owner_constants import OWNER_OF_OBJECT_RELATION_TYPE, OWNER_RELATION_TYPE
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.user import User


class Owner(GraphSerializable, TableSerializable):
    LABELS_PERMITTED_TO_HAVE_OWNER = ['Table', 'Dashboard', 'Feature']

    def __init__(self,
                 start_label: str,
                 start_key: str,
                 owner_emails: Union[List, str],
                 ) -> None:
        if start_label not in Owner.LABELS_PERMITTED_TO_HAVE_OWNER:
            raise Exception(f'owners for {start_label} are not supported')
        self.start_label = start_label
        self.start_key = start_key
        if isinstance(owner_emails, str):
            owner_emails = owner_emails.split(',')
        self.owner_emails = [email.strip().lower() for email in owner_emails]

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()

    def __repr__(self) -> str:
        return f'Owner({self.start_label!r}, {self.start_key!r}, {self.owner_emails!r})'

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        for email in self.owner_emails:
            if email:
                yield GraphNode(
                    key=User.get_user_model_key(email=email),
                    label=User.USER_NODE_LABEL,
                    attributes={
                        User.USER_NODE_EMAIL: email,
                    }
                )

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        for email in self.owner_emails:
            if email:
                yield GraphRelationship(
                    start_label=self.start_label,
                    start_key=self.start_key,
                    end_label=User.USER_NODE_LABEL,
                    end_key=User.get_user_model_key(email=email),
                    type=OWNER_RELATION_TYPE,
                    reverse_type=OWNER_OF_OBJECT_RELATION_TYPE,
                    attributes={}
                )

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        for email in self.owner_emails:
            if email:
                user_record = RDSUser(
                    rk=User.get_user_model_key(email=email),
                    email=email
                )
                yield user_record

                if self.start_label == TableMetadata.TABLE_NODE_LABEL:
                    yield RDSTableOwner(
                        table_rk=self.start_key,
                        user_rk=User.get_user_model_key(email=email),
                    )
                elif self.start_label == DashboardMetadata.DASHBOARD_NODE_LABEL:
                    yield RDSDashboardOwner(
                        dashboard_rk=self.start_key,
                        user_rk=User.get_user_model_key(email=email)
                    )
                else:
                    raise Exception(f'{self.start_label}<>Owner relationship is not table serializable')



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator, Union

from amundsen_rds.models import RDSModel
from amundsen_rds.models.updated_timestamp import UpdatedTimestamp as RDSUpdatedTimestamp

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable


class ESLastUpdated(GraphSerializable, TableSerializable):
    """
    Data model to keep track the last updated timestamp for
    datastore and es.
    """

    LABEL = 'Updatedtimestamp'
    KEY = 'amundsen_updated_timestamp'
    LATEST_TIMESTAMP = 'latest_timestamp'

    def __init__(self,
                 timestamp: int,
                 ) -> None:
        """
        :param timestamp: epoch for latest updated timestamp for neo4j an es
        """
        self.timestamp = timestamp
        self._node_iter = self._create_node_iterator()
        self._rel_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        """
        Will create an orphan node for last updated timestamp.
        """
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create an es_updated_timestamp node
        """
        node = GraphNode(
            key=ESLastUpdated.KEY,
            label=ESLastUpdated.LABEL,
            attributes={
                ESLastUpdated.LATEST_TIMESTAMP: self.timestamp
            }
        )
        yield node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        return
        yield

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = RDSUpdatedTimestamp(rk=ESLastUpdated.KEY,
                                     latest_timestamp=self.timestamp)
        yield record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (  # noqa: F401
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401

from databuilder.extractor.base_postgres_metadata_extractor import BasePostgresMetadataExtractor


class RedshiftMetadataExtractor(BasePostgresMetadataExtractor):
    """
    Extracts Redshift table and column metadata from underlying meta store database using SQLAlchemyExtractor

    This differs from the PostgresMetadataExtractor because in order to support Redshift's late binding views,
    we need to join the INFORMATION_SCHEMA data against the function PG_GET_LATE_BINDING_VIEW_COLS().
    """

    def get_sql_statement(self, use_catalog_as_cluster_name: bool, where_clause_suffix: str) -> str:
        if use_catalog_as_cluster_name:
            cluster_source = "CURRENT_DATABASE()"
        else:
            cluster_source = f"'{self._cluster}'"

        return """
        SELECT
            *
        FROM (
            SELECT
              {cluster_source} as cluster,
              c.table_schema as schema,
              c.table_name as name,
              pgtd.description as description,
              c.column_name as col_name,
              c.data_type as col_type,
              pgcd.description as col_description,
              ordinal_position as col_sort_order
            FROM INFORMATION_SCHEMA.COLUMNS c
            INNER JOIN
              pg_catalog.pg_statio_all_tables as st on c.table_schema=st.schemaname and c.table_name=st.relname
            LEFT JOIN
              pg_catalog.pg_description pgcd on pgcd.objoid=st.relid and pgcd.objsubid=c.ordinal_position
            LEFT JOIN
              pg_catalog.pg_description pgtd on pgtd.objoid=st.relid and pgtd.objsubid=0

            UNION

            SELECT
              {cluster_source} as cluster,
              view_schema as schema,
              view_name as name,
              NULL as description,
              column_name as col_name,
              data_type as col_type,
              NULL as col_description,
              ordinal_position as col_sort_order
            FROM
                PG_GET_LATE_BINDING_VIEW_COLS()
                    COLS(view_schema NAME, view_name NAME, column_name NAME, data_type VARCHAR, ordinal_position INT)

            UNION

            SELECT
              {cluster_source} AS cluster,
              schemaname AS schema,
              tablename AS name,
              NULL AS description,
              columnname AS col_name,
              external_type AS col_type,
              NULL AS col_description,
              columnnum AS col_sort_order
            FROM svv_external_columns
        )

        {where_clause_suffix}
        ORDER by cluster, schema, name, col_sort_order ;
        """.format(
            cluster_source=cluster_source,
            where_clause_suffix=where_clause_suffix,
        )

    def get_scope(self) -> str:
        return 'extractor.redshift_metadata'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import datetime
import logging
import os
import time
from os import listdir
from os.path import isfile, join
from typing import List, Tuple

from amundsen_gremlin.neptune_bulk_loader.api import NeptuneBulkLoaderApi, NeptuneBulkLoaderLoadStatusErrorLogEntry
from boto3.session import Session
from pyhocon import ConfigTree

from databuilder.publisher.base_publisher import Publisher

LOGGER = logging.getLogger(__name__)


class NeptuneCSVPublisher(Publisher):
    """
    This Publisher takes two folders for input and publishes to Neptune.
    One folder will contain CSV file(s) for Node where the other folder will contain CSV file(s) for Relationship.

    This publisher uses the bulk api found in
    https://github.com/amundsen-io/amundsengremlin/blob/master/amundsen_gremlin/neptune_bulk_loader/api.py

    which is a client for the the api found
    https://docs.aws.amazon.com/neptune/latest/userguide/bulk-load.html
    https://docs.aws.amazon.com/neptune/latest/userguide/bulk-load-tutorial-format-gremlin.html
    """

    # A directory that contains CSV files for nodes
    NODE_FILES_DIR = 'node_files_directory'
    # A directory that contains CSV files for relationships
    RELATION_FILES_DIR = 'relation_files_directory'

    # --- AWS CONFIGURATION ---
    # S3 bucket to upload files to
    AWS_S3_BUCKET_NAME = 'bucket_name'
    # S3 location where amundsen data can be exported to and Neptune can access
    AWS_BASE_S3_DATA_PATH = 'base_amundsen_data_path'

    NEPTUNE_HOST = 'neptune_host'

    # AWS CONFIGURATION
    AWS_REGION = 'aws_region'
    AWS_ACCESS_KEY = 'aws_access_key'
    AWS_SECRET_ACCESS_KEY = 'aws_secret_access_key'
    AWS_SESSION_TOKEN = 'aws_session_token'
    AWS_IAM_ROLE_NAME = 'aws_iam_role_name'

    FAIL_ON_ERROR = "fail_on_error"
    STATUS_POLLING_PERIOD = "status_polling_period"

    def __init__(self) -> None:
        super(NeptuneCSVPublisher, self).__init__()

    def init(self, conf: ConfigTree) -> None:
        self._boto_session = Session(
            aws_access_key_id=conf.get_string(NeptuneCSVPublisher.AWS_ACCESS_KEY, default=None),
            aws_secret_access_key=conf.get_string(NeptuneCSVPublisher.AWS_SECRET_ACCESS_KEY, default=None),
            aws_session_token=conf.get_string(NeptuneCSVPublisher.AWS_SESSION_TOKEN, default=None),
            region_name=conf.get_string(NeptuneCSVPublisher.AWS_REGION, default=None)
        )

        self.node_files_dir = conf.get_string(NeptuneCSVPublisher.NODE_FILES_DIR)
        self.relation_files_dir = conf.get_string(NeptuneCSVPublisher.RELATION_FILES_DIR)

        self._neptune_host = conf.get_string(NeptuneCSVPublisher.NEPTUNE_HOST)

        neptune_bulk_endpoint_uri = "wss://{host}/gremlin".format(
            host=self._neptune_host
        )

        self.bucket_name = conf.get_string(NeptuneCSVPublisher.AWS_S3_BUCKET_NAME)

        self.neptune_api_client = NeptuneBulkLoaderApi(
            session=self._boto_session,
            endpoint_uri=neptune_bulk_endpoint_uri,
            s3_bucket_name=self.bucket_name,
            iam_role_name=conf.get_string(NeptuneCSVPublisher.AWS_IAM_ROLE_NAME, default=None)
        )
        self.base_amundsen_data_path = conf.get_string(NeptuneCSVPublisher.AWS_BASE_S3_DATA_PATH)
        self.fail_on_error = conf.get_bool(NeptuneCSVPublisher.FAIL_ON_ERROR, default=False)
        self.status_polling_period = conf.get_int(NeptuneCSVPublisher.STATUS_POLLING_PERIOD, default=5)

    def publish_impl(self) -> None:
        if not self._is_upload_required():
            return

        datetime_portion = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        s3_folder_location = "{base_directory}/{datetime_portion}".format(
            base_directory=self.base_amundsen_data_path,
            datetime_portion=datetime_portion,
        )

        self.upload_files(s3_folder_location)

        bulk_upload_response = self.neptune_api_client.load(
            s3_object_key=s3_folder_location,
            failOnError=self.fail_on_error
        )

        try:
            load_id = bulk_upload_response['payload']['loadId']
        except KeyError:
            raise Exception("Failed to load csv. Response: {0}".format(str(bulk_upload_response)))

        load_status = "LOAD_NOT_STARTED"
        all_errors: List[NeptuneBulkLoaderLoadStatusErrorLogEntry] = []
        while load_status in ("LOAD_IN_PROGRESS", "LOAD_NOT_STARTED", "LOAD_IN_QUEUE"):
            time.sleep(self.status_polling_period)
            load_status, errors = self._poll_status(load_id)
            all_errors.extend(errors)

        for error in all_errors:
            exception_message = """
            Error Code: {error_code}
            Error Message: {error_message}
            Failed File: {s3_path}
            """.format(
                error_code=error.get('errorCode'),
                error_message=error.get('errorMessage'),
                s3_path=error.get('fileName')
            )
            LOGGER.exception(exception_message)

    def _poll_status(self, load_id: str) -> Tuple[str, List[NeptuneBulkLoaderLoadStatusErrorLogEntry]]:
        load_status_response = self.neptune_api_client.load_status(
            load_id=load_id,
            errors=True
        )
        load_status_payload = load_status_response.get('payload', {})
        try:
            load_status = load_status_payload['overallStatus']['status']
        except KeyError:
            raise Exception("Failed to check status of {0} response: {1}".format(
                str(load_id),
                repr(load_status_response)
            ))
        return load_status, load_status_payload.get('errors', {}).get('errorLogs', [])

    def _get_file_paths(self) -> List[str]:
        node_names = [
            join(self.node_files_dir, f) for f in listdir(self.node_files_dir)
            if isfile(join(self.node_files_dir, f))
        ]
        edge_names = [
            join(self.relation_files_dir, f) for f in listdir(self.relation_files_dir)
            if isfile(join(self.relation_files_dir, f))
        ]
        return node_names + edge_names

    def _is_upload_required(self) -> bool:
        file_names = self._get_file_paths()
        return len(file_names) > 0

    def upload_files(self, s3_folder_location: str) -> None:
        file_paths = self._get_file_paths()
        for file_location in file_paths:
            with open(file_location, 'rb') as file_csv:
                file_name = os.path.basename(file_location)
                s3_object_key = "{s3_folder_location}/{file_name}".format(
                    s3_folder_location=s3_folder_location,
                    file_name=file_name
                )
                self.neptune_api_client.upload(
                    f=file_csv,
                    s3_object_key=s3_object_key
                )

    def get_scope(self) -> str:
        return 'publisher.neptune_csv_publisher'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Union

from databuilder.models.owner import Owner
from databuilder.models.table_metadata import TableMetadata


class TableOwner(Owner):
    """
    Table owner model.
    """

    def __init__(self,
                 db_name: str,
                 schema: str,
                 table_name: str,
                 owners: Union[List, str],
                 cluster: str = 'gold',
                 ) -> None:
        self.start_label = TableMetadata.TABLE_NODE_LABEL
        self.start_key = f'{db_name}://{cluster}.{schema}/{table_name}'

        Owner.__init__(
            self,
            start_label=self.start_label,
            start_key=self.start_key,
            owner_emails=owners,
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Any

from pyhocon import ConfigTree

from databuilder import Scoped


class Loader(Scoped):
    """
    A loader loads to the destination or to the staging area
    """
    @abc.abstractmethod
    def init(self, conf: ConfigTree) -> None:
        pass

    @abc.abstractmethod
    def load(self, record: Any) -> None:
        pass

    def get_scope(self) -> str:
        return 'loader'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Union

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship


class AtlasSerializable(object, metaclass=abc.ABCMeta):
    """
    A Serializable abstract class asks subclass to implement next node or
    next relation in dict form so that it can be serialized to CSV file.

    Any model class that needs to be pushed to a atlas should inherit this class.
    """

    @abc.abstractmethod
    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        """
        Creates Atlas entity the process that consumes this class takes the output
        serializes to atlas entity

        :return: a Atlas entity or None if no more records to serialize
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        """
        Creates AtlasRelationship the process that consumes this class takes the output
        serializes to the desired graph database.

        :return: a AtlasRelationship or None if no more record to serialize
        """
        raise NotImplementedError

    def _validate_atlas_entity(self, entity: AtlasEntity) -> None:
        entity_type, operation, relation, attributes = entity

        if entity_type is None:
            raise ValueError('Required header missing: entityType')

        if operation is None:
            raise ValueError('Required header missing: operation')

        if attributes is None:
            raise ValueError('Required header missing: attributes')

        if 'qualifiedName' not in attributes:
            raise ValueError(f'Attribute qualifiedName is missing')

    def _validate_atlas_relation(self, relation: AtlasRelationship) -> None:
        relation_type, entity_type_1, qualified_name_1, entity_type_2, qualified_name_2, _ = relation

        if relation_type is None:
            raise ValueError(f'Required header missing. Missing: {AtlasRelationship.relationshipType}')

        if entity_type_1 is None:
            raise ValueError(f'Required header missing. Missing: {AtlasRelationship.entityType1}')

        if qualified_name_1 is None:
            raise ValueError(f'Required header missing. Missing: {AtlasRelationship.entityQualifiedName1}')

        if entity_type_2 is None:
            raise ValueError(f'Required header missing. Missing: {AtlasRelationship.entityType2}')

        if qualified_name_2 is None:
            raise ValueError(f'Required header missing. Missing: {AtlasRelationship.entityQualifiedName2}')

    def next_atlas_entity(self) -> Union[AtlasEntity, None]:
        entity_dict = self.create_next_atlas_entity()
        if not entity_dict:
            return None

        self._validate_atlas_entity(entity_dict)
        return entity_dict

    def next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        relation_dict = self.create_next_atlas_relation()
        if not relation_dict:
            return None

        self._validate_atlas_relation(relation_dict)
        return relation_dict



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

GraphRelationship = namedtuple(
    'GraphRelationship',
    [
        'start_label',
        'end_label',
        'start_key',
        'end_key',
        'type',
        'reverse_type',
        'attributes'
    ]
)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple
from itertools import groupby
from typing import (
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


class Db2MetadataExtractor(Extractor):
    """
    Extracts Db2 table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """
    # SELECT statement from Db2 SYSIBM to extract table and column metadata
    SQL_STATEMENT = """
    SELECT
      {cluster_source} as cluster, c.TABSCHEMA as schema, c.TABNAME as name, t.REMARKS as description,
      c.COLNAME as col_name,
      CASE WHEN c.TYPENAME='VARCHAR' OR c.TYPENAME='CHARACTER' THEN
      TRIM (TRAILING FROM c.TYPENAME) concat '(' concat c.LENGTH concat ')'
      WHEN c.TYPENAME='DECIMAL' THEN
      TRIM (TRAILING FROM c.TYPENAME) concat '(' concat c.LENGTH concat ',' concat c.SCALE concat ')'
      ELSE TRIM (TRAILING FROM c.TYPENAME) END as col_type,
      c.REMARKS as col_description, c.COLNO as col_sort_order
    FROM SYSCAT.COLUMNS c
    INNER JOIN
      SYSCAT.TABLES as t on c.TABSCHEMA=t.TABSCHEMA and c.TABNAME=t.TABNAME
    {where_clause_suffix}
    ORDER by cluster, schema, name, col_sort_order ;
    """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster_key'
    DATABASE_KEY = 'database_key'

    # Default values
    DEFAULT_CLUSTER_NAME = 'master'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {WHERE_CLAUSE_SUFFIX_KEY: ' ', CLUSTER_KEY: DEFAULT_CLUSTER_NAME}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(Db2MetadataExtractor.DEFAULT_CONFIG)
        self._cluster = conf.get_string(Db2MetadataExtractor.CLUSTER_KEY)

        cluster_source = f"'{self._cluster}'"

        self._database = conf.get_string(Db2MetadataExtractor.DATABASE_KEY, default='db2')

        self.sql_stmt = Db2MetadataExtractor.SQL_STATEMENT.format(
            where_clause_suffix=conf.get_string(Db2MetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY),
            cluster_source=cluster_source
        )

        self._alchemy_extractor = SQLAlchemyExtractor()
        sql_alch_conf = Scoped.get_scoped_conf(conf, self._alchemy_extractor.get_scope())\
            .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt}))

        self.sql_stmt = sql_alch_conf.get_string(SQLAlchemyExtractor.EXTRACT_SQL)

        LOGGER.info('SQL for Db2 metadata: %s', self.sql_stmt)

        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.db2_metadata'

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        """
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        """
        for key, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                columns.append(ColumnMetadata(row['col_name'], row['col_description'],
                                              row['col_type'], row['col_sort_order']))

            yield TableMetadata(self._database, last_row['cluster'],
                                last_row['schema'],
                                last_row['name'],
                                last_row['description'],
                                columns)

    def _get_raw_extract_iter(self) -> Iterator[Dict[str, Any]]:
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()

    def _get_table_key(self, row: Dict[str, Any]) -> Union[TableKey, None]:
        """
        Table key consists of schema and table name
        :param row:
        :return:
        """
        if row:
            return TableKey(schema=row['schema'], table_name=row['name'])

        return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, List, Optional, Tuple,
)

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.utils.atlas import AtlasSerializedEntityFields, AtlasSerializedRelationshipFields


def serialize_entity(entity: Optional[AtlasEntity]) -> Dict[str, Any]:
    if entity is None:
        return {}

    entity_dict = {
        AtlasSerializedEntityFields.type_name: entity.typeName,
        AtlasSerializedEntityFields.operation: entity.operation,
        AtlasSerializedEntityFields.relationships: entity.relationships
    }
    for key, value in entity.attributes.items():
        entity_dict[key] = value
    return entity_dict


def serialize_relationship(relationship: Optional[AtlasRelationship]) -> Dict[str, Any]:
    if relationship is None:
        return {}

    relationship_dict = {
        AtlasSerializedRelationshipFields.relation_type: relationship.relationshipType,
        AtlasSerializedRelationshipFields.entity_type_1: relationship.entityType1,
        AtlasSerializedRelationshipFields.qualified_name_1: relationship.entityQualifiedName1,
        AtlasSerializedRelationshipFields.entity_type_2: relationship.entityType2,
        AtlasSerializedRelationshipFields.qualified_name_2: relationship.entityQualifiedName2,
    }
    for key, value in relationship.attributes.items():
        relationship_dict[key] = value

    return relationship_dict


def get_entity_attrs(attrs_mapping: List[Tuple[Any, Any]]) -> Dict:
    entity_attrs = {}
    for attr in attrs_mapping:
        attr_key, attr_value = attr
        entity_attrs[attr_key] = attr_value
    return entity_attrs


def add_entity_relationship(
    relationship_list: List[str], relation_attribute: str,
    relation_entity_type: str, related_object_qualified_name: str,
) -> List[str]:
    """
    relationship in form 'relation_attribute#relation_entity_type#qualified_name_of_related_object
    """
    relationship_list.append(AtlasSerializedEntityFields.relationships_kv_separator.join(
        (relation_attribute, relation_entity_type, related_object_qualified_name),
    ))
    return relationship_list


def get_entity_relationships(relationship_list: List[str]) -> str:
    return AtlasSerializedEntityFields \
        .relationships_separator \
        .join(relationship_list)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Dict, Iterator, List,
)

from pyhocon import ConfigFactory, ConfigTree

import databuilder.extractor.dashboard.tableau.tableau_dashboard_constants as const
from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.tableau.tableau_dashboard_utils import (
    TableauDashboardUtils, TableauGraphQLApiExtractor,
)
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT
from databuilder.transformer.base_transformer import ChainedTransformer, Transformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME, TimestampStringToEpoch

LOGGER = logging.getLogger(__name__)


class TableauGraphQLApiMetadataExtractor(TableauGraphQLApiExtractor):
    """
    Implements the extraction-time logic for parsing the GraphQL result and transforming into a dict
    that fills the DashboardMetadata model. Allows workbooks to be exlcuded based on their project.
    """

    CLUSTER = const.CLUSTER
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    SITE_NAME = const.SITE_NAME
    TABLEAU_BASE_URL = const.TABLEAU_BASE_URL

    def execute(self) -> Iterator[Dict[str, Any]]:
        response = self.execute_query()

        workbooks_data = [workbook for workbook in response['workbooks']
                          if workbook['projectName'] not in
                          self._conf.get_list(TableauGraphQLApiMetadataExtractor.EXCLUDED_PROJECTS, [])]
        base_url = self._conf.get(TableauGraphQLApiMetadataExtractor.TABLEAU_BASE_URL)
        site_name = self._conf.get_string(TableauGraphQLApiMetadataExtractor.SITE_NAME, '')
        site_url_path = ''
        if site_name != '':
            site_url_path = f'/site/{site_name}'
        for workbook in workbooks_data:
            if None in (workbook['projectName'], workbook['name']):
                LOGGER.warning(f'Ignoring workbook (ID:{workbook["vizportalUrlId"]}) ' +
                               f'in project (ID:{workbook["projectVizportalUrlId"]}) because of a lack of permission')
                continue
            data = {
                'dashboard_group': workbook['projectName'],
                'dashboard_name': TableauDashboardUtils.sanitize_workbook_name(workbook['name']),
                'description': workbook.get('description', ''),
                'created_timestamp': workbook['createdAt'],
                'dashboard_group_url': f'{base_url}/#{site_url_path}/projects/{workbook["projectVizportalUrlId"]}',
                'dashboard_url': f'{base_url}/#{site_url_path}/workbooks/{workbook["vizportalUrlId"]}/views',
                'cluster': self._conf.get_string(TableauGraphQLApiMetadataExtractor.CLUSTER)
            }
            yield data


class TableauDashboardExtractor(Extractor):
    """
    Extracts core metadata about Tableau "dashboards".
    For the purposes of this extractor, Tableau "workbooks" are mapped to Amundsen dashboards, and the
    top-level project in which these workbooks preside is the dashboard group. The metadata it gathers is:
        Dashboard name (Workbook name)
        Dashboard description (Workbook description)
        Dashboard creation timestamp (Workbook creationstamp)
        Dashboard group name (Workbook top-level folder name)
    Uses the Metadata API: https://help.tableau.com/current/api/metadata_api/en-us/index.html
    """

    API_BASE_URL = const.API_BASE_URL
    API_VERSION = const.API_VERSION
    CLUSTER = const.CLUSTER
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    SITE_NAME = const.SITE_NAME
    TABLEAU_BASE_URL = const.TABLEAU_BASE_URL
    TABLEAU_ACCESS_TOKEN_NAME = const.TABLEAU_ACCESS_TOKEN_NAME
    TABLEAU_ACCESS_TOKEN_SECRET = const.TABLEAU_ACCESS_TOKEN_SECRET
    VERIFY_REQUEST = const.VERIFY_REQUEST

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self.query = """query {
            workbooks {
                id
                name
                createdAt
                description
                projectName
                projectVizportalUrlId
                vizportalUrlId
            }
        }"""

        self._extractor = self._build_extractor()

        transformers: List[Transformer] = []
        timestamp_str_to_epoch_transformer = TimestampStringToEpoch()
        timestamp_str_to_epoch_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, timestamp_str_to_epoch_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict({FIELD_NAME: 'created_timestamp', })))
        transformers.append(timestamp_str_to_epoch_transformer)

        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_metadata.DashboardMetadata'})))
        transformers.append(dict_to_model_transformer)
        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()
        if not record:
            return None

        return next(self._transformer.transform(record=record), None)

    def get_scope(self) -> str:
        return 'extractor.tableau_dashboard_metadata'

    def _build_extractor(self) -> TableauGraphQLApiMetadataExtractor:
        """
        Builds a TableauGraphQLApiMetadataExtractor. All data required can be retrieved with a single GraphQL call.
        :return: A TableauGraphQLApiMetadataExtractor that provides core dashboard metadata.
        """
        extractor = TableauGraphQLApiMetadataExtractor()
        tableau_extractor_conf = Scoped.get_scoped_conf(self._conf, extractor.get_scope()) \
            .with_fallback(self._conf) \
            .with_fallback(ConfigFactory.from_dict({TableauGraphQLApiExtractor.QUERY: self.query,
                                                    STATIC_RECORD_DICT: {'product': 'tableau'}}))
        extractor.init(conf=tableau_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


from typing import (
    Any, Dict, Iterator, List, Tuple, Union,
)

from databuilder.extractor.dashboard.apache_superset.apache_superset_extractor import (
    ApacheSupersetBaseExtractor, type_fields_mapping,
)
from databuilder.models.dashboard.dashboard_chart import DashboardChart
from databuilder.models.dashboard.dashboard_query import DashboardQuery


class ApacheSupersetChartExtractor(ApacheSupersetBaseExtractor):
    def chart_field_mappings(self) -> type_fields_mapping:
        result = [
            ('chart_id', 'id', lambda x: str(x), ''),
            ('chart_name', 'slice_name', None, ''),
            ('chart_type', 'viz_type', None, ''),
            ('chart_url', 'url', None, ''),  # currently not available in superset chart api
        ]

        return result

    def _get_extract_iter(self) -> Iterator[Union[DashboardQuery, DashboardChart, None]]:
        ids = self._get_dashboard_ids()

        data = [self._get_dashboard_details(i) for i in ids]

        for entry in data:
            dashboard_id, dashboard_data = entry

            # Since Apache Superset doesn't support dashboard <> query <> chart relation we create a dummy 'bridge'
            # query node so we can connect charts to a dashboard
            dashboard_query_data = dict(dashboard_id=dashboard_id,
                                        query_name='default',
                                        query_id=self.dummy_query_id,
                                        url='',
                                        query_text='')
            dashboard_query_data.update(**self.common_params)

            yield DashboardQuery(**dashboard_query_data)

            charts = [s.get('__Slice__') for s in dashboard_data.get('slices', [])]

            for chart in charts:
                if chart:
                    dashboard_chart_data = self.map_fields(chart, self.chart_field_mappings())
                    dashboard_chart_data.update(**{**dict(dashboard_id=dashboard_id, query_id=self.dummy_query_id),
                                                   **self.common_params})

                    yield DashboardChart(**dashboard_chart_data)

    def _get_dashboard_ids(self) -> List[str]:
        url = self.build_full_url(f'api/v1/dashboard')

        data = self.execute_query(url)

        return data.get('ids', [])

    def _get_dashboard_details(self, dashboard_id: str) -> Tuple[str, Dict[str, Any]]:
        url = self.build_full_url(f'api/v1/dashboard/export?q=[{dashboard_id}]')

        _data = self.execute_query(url)

        dashboard_data = _data.get('dashboards', [])[0].get('__Dashboard__', dict())

        data = dashboard_id, dashboard_data

        return data

    @property
    def dummy_query_id(self) -> str:
        return '-1'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import concurrent.futures
import logging
from collections import namedtuple
from datetime import datetime
from typing import (  # noqa: F401
    Dict, Iterator, List, Optional, Union,
)

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401
from pyspark.sql import SparkSession
from pyspark.sql.catalog import Table
from pyspark.sql.utils import AnalysisException

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_last_updated import TableLastUpdated
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


# TODO once column tags work properly, consider deprecating this for TableMetadata directly
class ScrapedColumnMetadata(object):
    def __init__(self, name: str, data_type: str, description: Optional[str], sort_order: int):
        self.name = name
        self.data_type = data_type
        self.description = description
        self.sort_order = sort_order
        self.is_partition = False
        self.attributes: Dict[str, str] = {}

    def set_is_partition(self, is_partition: bool) -> None:
        self.is_partition = is_partition

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ScrapedColumnMetadata):
            return False
        return (self.name == other.name and
                self.data_type == other.data_type and
                self.description == other.description and
                self.sort_order == other.sort_order and
                self.is_partition == other.is_partition and
                self.attributes == other.attributes)

    def __repr__(self) -> str:
        return f'{self.name}:{self.data_type}'


# TODO consider deprecating this for using TableMetadata directly
class ScrapedTableMetadata(object):
    LAST_MODIFIED_KEY = 'lastModified'
    DESCRIPTION_KEY = 'description'
    TABLE_FORMAT_KEY = 'format'

    def __init__(self, schema: str, table: str):
        self.schema: str = schema
        self.table: str = table
        self.table_detail: Optional[Dict] = None
        self.view_detail: Optional[Dict] = None
        self.is_view: bool = False
        self.failed_to_scrape: bool = False
        self.columns: Optional[List[ScrapedColumnMetadata]] = None

    def set_table_detail(self, table_detail: Dict) -> None:
        self.table_detail = table_detail
        self.is_view = False
        self.failed_to_scrape = False

    def set_view_detail(self, view_detail: Dict) -> None:
        self.view_detail = view_detail
        self.is_view = True
        self.failed_to_scrape = False

    def get_details(self) -> Optional[Dict]:
        if self.is_view:
            return self.view_detail
        else:
            return self.table_detail

    def get_full_table_name(self) -> str:
        return self.schema + "." + self.table

    def set_failed_to_scrape(self) -> None:
        self.failed_to_scrape = True

    def set_columns(self, column_list: List[ScrapedColumnMetadata]) -> None:
        self.columns = column_list

    def get_last_modified(self) -> Optional[datetime]:
        details = self.get_details()
        if details and self.LAST_MODIFIED_KEY in details:
            return details[self.LAST_MODIFIED_KEY]
        else:
            return None

    def get_table_description(self) -> Optional[str]:
        details = self.get_details()
        if details and self.DESCRIPTION_KEY in details:
            return details[self.DESCRIPTION_KEY]
        else:
            return None

    def is_delta_table(self) -> bool:
        details = self.get_details()
        if details and self.TABLE_FORMAT_KEY in details:
            return details[self.TABLE_FORMAT_KEY].lower() == 'delta'
        else:
            return False

    def __repr__(self) -> str:
        return f'{self.schema}.{self.table}'


class DeltaLakeMetadataExtractor(Extractor):
    """
    Extracts Delta Lake Metadata.
    This requires a spark session to run that has a hive metastore populated with all of the delta tables
    that you are interested in.
    """
    # CONFIG KEYS
    DATABASE_KEY = "database"
    # If you want to exclude specific schemas
    EXCLUDE_LIST_SCHEMAS_KEY = "exclude_list"
    # If you want to only include specific schemas
    SCHEMA_LIST_KEY = "schema_list"
    CLUSTER_KEY = "cluster"
    # By default, this will only process and emit delta-lake tables, but it can support all hive table types.
    DELTA_TABLES_ONLY = "delta_tables_only"
    DEFAULT_CONFIG = ConfigFactory.from_dict({DATABASE_KEY: "delta",
                                              EXCLUDE_LIST_SCHEMAS_KEY: [],
                                              SCHEMA_LIST_KEY: [],
                                              DELTA_TABLES_ONLY: True})
    PARTITION_COLUMN_TAG = 'is_partition'

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf.with_fallback(DeltaLakeMetadataExtractor.DEFAULT_CONFIG)
        self._extract_iter = None  # type: Union[None, Iterator]
        self._cluster = self.conf.get_string(DeltaLakeMetadataExtractor.CLUSTER_KEY)
        self._db = self.conf.get_string(DeltaLakeMetadataExtractor.DATABASE_KEY)
        self.exclude_list = self.conf.get_list(DeltaLakeMetadataExtractor.EXCLUDE_LIST_SCHEMAS_KEY)
        self.schema_list = self.conf.get_list(DeltaLakeMetadataExtractor.SCHEMA_LIST_KEY)
        self.delta_tables_only = self.conf.get_bool(DeltaLakeMetadataExtractor.DELTA_TABLES_ONLY)

    def set_spark(self, spark: SparkSession) -> None:
        self.spark = spark

    def extract(self) -> Union[TableMetadata, TableLastUpdated, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.delta_lake_table_metadata'

    def _get_extract_iter(self) -> Iterator[Union[TableMetadata, TableLastUpdated, None]]:
        """
        Given either a list of schemas, or a list of exclude schemas,
        it will query hive metastore and then access delta log
        to get all of the metadata for your delta tables. It will produce:
         - table and column metadata
         - last updated information
        """
        if self.schema_list:
            LOGGER.info("working on %s", self.schema_list)
            tables = self.get_all_tables(self.schema_list)
        else:
            LOGGER.info("fetching all schemas")
            LOGGER.info("Excluding: %s", self.exclude_list)
            schemas = self.get_schemas(self.exclude_list)
            LOGGER.info("working on %s", schemas)
            tables = self.get_all_tables(schemas)
        # TODO add the programmatic information as well?
        # TODO add watermarks
        scraped_tables = self.scrape_all_tables(tables)
        for scraped_table in scraped_tables:
            if not scraped_table:
                continue
            if self.delta_tables_only and not scraped_table.is_delta_table():
                LOGGER.info("Skipping none delta table %s", scraped_table.table)
                continue
            else:
                yield self.create_table_metadata(scraped_table)
                last_updated = self.create_table_last_updated(scraped_table)
                if last_updated:
                    yield last_updated

    def get_schemas(self, exclude_list: List[str]) -> List[str]:
        '''Returns all schemas.'''
        schemas = self.spark.catalog.listDatabases()
        ret = []
        for schema in schemas:
            if schema.name not in exclude_list:
                ret.append(schema.name)
        return ret

    def get_all_tables(self, schemas: List[str]) -> List[Table]:
        '''Returns all tables.'''
        ret = []
        for schema in schemas:
            ret.extend(self.get_tables_for_schema(schema))
        return ret

    def get_tables_for_schema(self, schema: str) -> List[Table]:
        '''Returns all tables for a specific schema.'''
        return self.spark.catalog.listTables(schema)

    def scrape_all_tables(self, tables: List[Table]) -> List[Optional[ScrapedTableMetadata]]:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.scrape_table, table) for table in tables]
        scraped_tables = [f.result() for f in futures]
        return scraped_tables

    def scrape_table(self, table: Table) -> Optional[ScrapedTableMetadata]:
        '''Takes a table object and creates a scraped table metadata object.'''
        met = ScrapedTableMetadata(schema=table.database, table=table.name)
        table_name = met.get_full_table_name()
        if table.tableType and table.tableType.lower() != 'view':
            table_detail = self.scrape_table_detail(table_name)
            if table_detail is None:
                LOGGER.error("Failed to parse table " + table_name)
                met.set_failed_to_scrape()
                return None
            else:
                LOGGER.info("Successfully parsed table " + table_name)
                met.set_table_detail(table_detail)
        else:
            view_detail = self.scrape_view_detail(table_name)
            if view_detail is None:
                LOGGER.error("Failed to parse view " + table_name)
                met.set_failed_to_scrape()
                return None
            else:
                LOGGER.info("Successfully parsed view " + table_name)
                met.set_view_detail(view_detail)
        columns = self.fetch_columns(met.schema, met.table)
        if not columns:
            LOGGER.error("Failed to parse columns for " + table_name)
            return None
        else:
            met.set_columns(columns)
            return met

    def scrape_table_detail(self, table_name: str) -> Optional[Dict]:
        try:
            table_details_df = self.spark.sql(f"describe detail {table_name}")
            table_detail = table_details_df.collect()[0]
            return table_detail.asDict()
        except Exception as e:
            LOGGER.error(e)
            return None

    def scrape_view_detail(self, view_name: str) -> Optional[Dict]:
        # TODO the blanket try catches need to be changed
        describeExtendedOutput = []
        try:
            describeExtendedOutput = self.spark.sql(f"describe extended {view_name}").collect()
        except Exception as e:
            LOGGER.error(e)
            return None
        view_detail = {}
        startAdding = False
        for row in describeExtendedOutput:
            row_dict = row.asDict()
            if startAdding:
                view_detail[row_dict['col_name']] = row_dict['data_type']
            if "# Detailed Table" in row_dict['col_name']:
                # Then start parsing
                startAdding = True
        return view_detail

    def fetch_columns(self, schema: str, table: str) -> List[ScrapedColumnMetadata]:
        '''This fetches delta table columns, which unfortunately
        in the general case cannot rely on spark.catalog.listColumns.'''
        raw_columns = []
        try:
            raw_columns = self.spark.sql(f"describe {schema}.{table}").collect()
        except AnalysisException as e:
            LOGGER.error(e)
            return raw_columns
        parsed_columns = {}
        partition_cols = False
        sort_order = 0
        for row in raw_columns:
            col_name = row['col_name']
            # NOTE: the behavior of describe has changed between spark 2 and spark 3
            if col_name == '' or '#' in col_name:
                partition_cols = True
                continue
            if not partition_cols:
                column = ScrapedColumnMetadata(
                    name=row['col_name'],
                    description=row['comment'] if row['comment'] else None,
                    data_type=row['data_type'],
                    sort_order=sort_order
                )
                parsed_columns[row['col_name']] = column
                sort_order += 1
            else:
                if row['data_type'] in parsed_columns:
                    LOGGER.debug(f"Adding partition column table for {row['data_type']}")
                    parsed_columns[row['data_type']].set_is_partition(True)
                elif row['col_name'] in parsed_columns:
                    LOGGER.debug(f"Adding partition column table for {row['col_name']}")
                    parsed_columns[row['col_name']].set_is_partition(True)
        return list(parsed_columns.values())

    def create_table_metadata(self, table: ScrapedTableMetadata) -> TableMetadata:
        '''Creates the amundsen table metadata object from the ScrapedTableMetadata object.'''
        amundsen_columns = []
        if table.columns:
            for column in table.columns:
                amundsen_columns.append(
                    ColumnMetadata(name=column.name,
                                   description=column.description,
                                   col_type=column.data_type,
                                   sort_order=column.sort_order)
                )
        description = table.get_table_description()
        return TableMetadata(self._db,
                             self._cluster,
                             table.schema,
                             table.table,
                             description,
                             amundsen_columns,
                             table.is_view)

    def create_table_last_updated(self, table: ScrapedTableMetadata) -> Optional[TableLastUpdated]:
        '''Creates the amundsen table last updated metadata object from the ScrapedTableMetadata object.'''
        last_modified = table.get_last_modified()
        if last_modified:
            return TableLastUpdated(table_name=table.table,
                                    last_updated_time_epoch=int(last_modified.timestamp()),
                                    schema=table.schema,
                                    db=self._db,
                                    cluster=self._cluster)
        else:
            return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Iterator

from pyhocon import ConfigTree

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.loader.base_loader import Loader
from databuilder.task.base_task import Task
from databuilder.transformer.base_transformer import NoopTransformer, Transformer
from databuilder.utils.closer import Closer

LOGGER = logging.getLogger(__name__)


class DefaultTask(Task):
    """
    A default task expecting to extract, transform and load.

    """

    # Determines the frequency of the log on task progress
    PROGRESS_REPORT_FREQUENCY = 'progress_report_frequency'

    def __init__(self,
                 extractor: Extractor,
                 loader: Loader,
                 transformer: Transformer = NoopTransformer()) -> None:
        self.extractor = extractor
        self.transformer = transformer
        self.loader = loader

        self._closer = Closer()
        self._closer.register(self.extractor.close)
        self._closer.register(self.transformer.close)
        self._closer.register(self.loader.close)

    def init(self, conf: ConfigTree) -> None:
        self._progress_report_frequency = \
            conf.get_int(f'{self.get_scope()}.{DefaultTask.PROGRESS_REPORT_FREQUENCY}', 500)

        self.extractor.init(Scoped.get_scoped_conf(conf, self.extractor.get_scope()))
        self.transformer.init(Scoped.get_scoped_conf(conf, self.transformer.get_scope()))
        self.loader.init(Scoped.get_scoped_conf(conf, self.loader.get_scope()))

    def run(self) -> None:
        """
        Runs a task
        """
        LOGGER.info('Running a task')
        try:
            record = self.extractor.extract()
            count = 0
            while record:
                record = self.transformer.transform(record)
                if not record:
                    # Move on if the transformer filtered the record out
                    record = self.extractor.extract()
                    continue

                # Support transformers which return one record, or yield multiple
                results = record if isinstance(record, Iterator) else [record]
                for result in results:
                    if result:
                        self.loader.load(result)
                        count += 1

                if count > 0 and count % self._progress_report_frequency == 0:
                    LOGGER.info(f'Extracted %i records so far', count)

                # Prepare the next record
                record = self.extractor.extract()
        finally:
            self._closer.close()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Union

from amundsen_rds.models import RDSModel


class TableSerializable(object, metaclass=abc.ABCMeta):
    """
    A Serializable abstract class asks subclass to implement next record
    in rds model instance form so that it can be serialized to CSV file.

    Any model class that needs to be pushed to a relational database should inherit this class.
    """

    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    def create_next_record(self) -> Union[RDSModel, None]:
        """
        Creates rds model instance.
        The process that consumes this class takes the output and serializes
        the record to the desired relational database.

        :return: a rds model instance or None if no more records to serialize
        """
        raise NotImplementedError

    def next_record(self) -> Union[RDSModel, None]:
        record = self.create_next_record()
        if not record:
            return None

        return record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import time
from typing import (
    Any, Dict, Set, Type,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.base import Base
from pyhocon import ConfigFactory, ConfigTree
from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker

from databuilder import Scoped
from databuilder.task.base_task import Task

LOGGER = logging.getLogger(__name__)


class MySQLStalenessRemovalTask(Task):
    """
    A Specific task to remove stale records in MySQL. It will use either the "published_tag" or
    "publisher_last_updated_epoch_ms" attribute assigned from the MySqlCsvPublisher. If "published tag" is different
    from the one it is getting from the config or if the "publisher_last_updated_epoch_ms" is less than the cutoff,
    it will regard the record as stale. Not all resources are being published by MySQLCSVPublisher and you can only
    set specific model table names to perform this deletion.
    Note: This task performs a cascade delete and will delete all the orphan records in the child tables of the stale
    records.

    """
    # Connection string
    CONN_STRING = "conn_string"
    # If its value is true, SQLAlchemy engine will log all statements
    ENGINE_ECHO = 'engine_echo'
    # Additional arguments used for engine
    CONNECT_ARGS = 'connect_args'

    TARGET_TABLES = "target_tables"
    DRY_RUN = "dry_run"
    # Staleness max percentage. Safety net to prevent majority of data being deleted.
    STALENESS_MAX_PCT = "staleness_max_pct"
    # Staleness max percentage per table. Safety net to prevent majority of data being deleted.
    STALENESS_PCT_MAX_DICT = "staleness_max_pct_dict"
    # If the published_tag value is different from the one in config, the record is determined to be stale
    PUBLISHED_TAG = "job_publish_tag"
    # Using this milliseconds and published timestamp to determine staleness
    MS_TO_EXPIRE = "milliseconds_to_expire"
    MIN_MS_TO_EXPIRE = "minimum_milliseconds_to_expire"

    _DEFAULT_CONFIG = ConfigFactory.from_dict({STALENESS_MAX_PCT: 5,
                                              TARGET_TABLES: [],
                                              STALENESS_PCT_MAX_DICT: {},
                                              MIN_MS_TO_EXPIRE: 86400000,
                                              DRY_RUN: False,
                                              ENGINE_ECHO: False})

    def get_scope(self) -> str:
        return 'task.mysql_remove_stale_data'

    def init(self, conf: ConfigTree) -> None:
        conf = Scoped.get_scoped_conf(conf, self.get_scope()) \
            .with_fallback(conf) \
            .with_fallback(MySQLStalenessRemovalTask._DEFAULT_CONFIG)
        self.target_tables = set(conf.get_list(MySQLStalenessRemovalTask.TARGET_TABLES))
        self.target_table_model_dict = self._get_target_table_model_dict(self.target_tables)
        self.dry_run = conf.get_bool(MySQLStalenessRemovalTask.DRY_RUN)
        self.staleness_max_pct = conf.get_int(MySQLStalenessRemovalTask.STALENESS_MAX_PCT)
        self.staleness_max_pct_dict = conf.get(MySQLStalenessRemovalTask.STALENESS_PCT_MAX_DICT)

        if MySQLStalenessRemovalTask.PUBLISHED_TAG in conf and MySQLStalenessRemovalTask.MS_TO_EXPIRE in conf:
            raise Exception(f'Cannot have both {MySQLStalenessRemovalTask.PUBLISHED_TAG} and '
                            f'{MySQLStalenessRemovalTask.MS_TO_EXPIRE} in job config')

        self.ms_to_expire = None
        if MySQLStalenessRemovalTask.MS_TO_EXPIRE in conf:
            self.ms_to_expire = conf.get_int(MySQLStalenessRemovalTask.MS_TO_EXPIRE)
            if self.ms_to_expire < conf.get_int(MySQLStalenessRemovalTask.MIN_MS_TO_EXPIRE):
                raise Exception(f'{MySQLStalenessRemovalTask.MS_TO_EXPIRE} is too small')
            self.marker = self.ms_to_expire
        else:
            self.marker = conf.get_string(MySQLStalenessRemovalTask.PUBLISHED_TAG)

        connect_args = {k: v for k, v in conf.get_config(MySQLStalenessRemovalTask.CONNECT_ARGS,
                                                         default=ConfigTree()).items()}
        self._engine = create_engine(conf.get_string(MySQLStalenessRemovalTask.CONN_STRING),
                                     echo=conf.get_bool(MySQLStalenessRemovalTask.ENGINE_ECHO),
                                     connect_args=connect_args)
        self._session_factory = sessionmaker(bind=self._engine)
        self._session = self._session_factory()

    def _get_target_table_model_dict(self, target_tables: Set[str]) -> Dict[str, Type[RDSModel]]:
        """
        Returns a dictionary with a table name to the corresponding RDS model class mapping.
        :param target_tables:
        :return:
        """
        target_table_model_dict: Dict[str, Type[RDSModel]] = {}
        for model in Base._decl_class_registry.values():
            if hasattr(model, '__tablename__') and model.__tablename__ in target_tables:
                target_table_model_dict[model.__tablename__] = model
        return target_table_model_dict

    def run(self) -> None:
        """
        For each table performs a safety check to make sure this operation does not delete more than a threshold
        where default threshold is 5%. Once it passes a safety check, it cascade deletes the stale records and all
        the referencing records. The target tables are traversed acc. to foreign key dependency (i.e in the order of
        child tables to parent tables) to try to keep the actual deleted record count consistent with the validation
        percentage (which reflects the staleness pct only from a single table and does not take into account the
        referenced tables data which will be deleted in a cascade delete)
        :return:
        """
        sorted_table_names = [table.name for table in Base.metadata.sorted_tables]
        sorted_target_tables = sorted(
            self.target_tables, key=lambda table: sorted_table_names.index(table), reverse=True)
        try:
            for table_name in sorted_target_tables:
                target_model_class = self.target_table_model_dict.get(table_name, None)
                if target_model_class:
                    staleness_pct = self._validate_record_staleness_pct(target_model_class=target_model_class)
                    if self.dry_run:
                        LOGGER.info('Skipping deleting records since it is a Dry Run.')
                        continue
                    if staleness_pct > 0:
                        self._delete_stale_records(target_model_class=target_model_class)
                else:
                    raise Exception(f'Failed to get corresponding model for {table_name}')
        except Exception as e:
            self._session.rollback()
            raise e
        finally:
            self._session.close()

    def _validate_record_staleness_pct(self, target_model_class: Type[RDSModel]) -> int:
        """
        Validation method.Focused on limit the risk of deleting records.
        - Check if deleted records for a table will be within 5% of the total records in the table.
        :param target_model_class:
        :return:
        """
        total_records_count = self._session.query(func.count('*')).select_from(target_model_class).scalar()
        stale_records_count = self._session.query(func.count('*')).select_from(target_model_class) \
            .filter(self._get_stale_records_filter_condition(target_model_class=target_model_class)).scalar()

        staleness_pct = 0
        target_table = target_model_class.__tablename__
        if stale_records_count:
            staleness_pct = stale_records_count * 100 / total_records_count
            threshold = self.staleness_max_pct_dict.get(target_table, self.staleness_max_pct)
            if staleness_pct >= threshold:
                raise Exception(f'Staleness percentage of {target_table} is {staleness_pct} %. '
                                f'Stopping due to over threshold {threshold} %')
            LOGGER.info(f'Will be deleting {stale_records_count} {target_table} '
                        f'record(s) or {staleness_pct}% of {target_table} data.')
        else:
            LOGGER.info(f'No stale records in {target_table}')
        return staleness_pct

    def _delete_stale_records(self, target_model_class: Type[RDSModel]) -> None:
        target_table = target_model_class.__tablename__
        try:
            deleted_records_count = self._session.query(target_model_class).filter(
                self._get_stale_records_filter_condition(target_model_class=target_model_class)).delete()
            self._session.commit()
            LOGGER.info(f'Deleted {deleted_records_count} record(s) of {target_table}')
        except Exception as e:
            LOGGER.exception(f'Failed to delete stale records for {target_table}')
            raise e

    def _get_stale_records_filter_condition(self, target_model_class: Type[RDSModel]) -> Any:
        """
        Return the appropriate stale records filter condition depending on which field is used to expire stale data.
        :param target_model_class:
        :return:
        """
        if self.ms_to_expire:
            current_time = int(time.time() * 1000)
            filter_condition = target_model_class.publisher_last_updated_epoch_ms < (
                current_time - self.marker)
        else:
            filter_condition = target_model_class.published_tag != self.marker
        return filter_condition



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from typing import Iterator, Union

import yaml
from feast import Client
from feast.feature_table import FeatureTable
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class FeastExtractor(Extractor):
    """
    Extracts feature tables from Feast Core service. Since Feast is
    a metadata store (and not the database itself), it maps the
    following atributes:

     * a database is name of feast project
     * table name is a name of the feature table
     * columns are features stored in the feature table
    """

    FEAST_SERVICE_CONFIG_KEY = "instance_name"
    FEAST_ENDPOINT_CONFIG_KEY = "endpoint"
    DESCRIBE_FEATURE_TABLES = "describe_feature_tables"
    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {FEAST_SERVICE_CONFIG_KEY: "main", DESCRIBE_FEATURE_TABLES: True}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(FeastExtractor.DEFAULT_CONFIG)
        self._feast_service = conf.get_string(FeastExtractor.FEAST_SERVICE_CONFIG_KEY)
        self._describe_feature_tables = conf.get_bool(
            FeastExtractor.DESCRIBE_FEATURE_TABLES
        )
        self._client = Client(
            core_url=conf.get_string(FeastExtractor.FEAST_ENDPOINT_CONFIG_KEY)
        )
        self._extract_iter: Union[None, Iterator] = None

    def get_scope(self) -> str:
        return "extractor.feast"

    def extract(self) -> Union[TableMetadata, None]:
        """
        For every feature table from Feast, a multiple objets are extracted:

        1. TableMetadata with feature table description
        2. Programmatic Description of the feature table, containing
           metadata - date of creation and labels
        3. Programmatic Description with Batch Source specification
        4. (if applicable) Programmatic Description with Stream Source
           specification
        """
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        for project in self._client.list_projects():
            for feature_table in self._client.list_feature_tables(project=project):
                yield from self._extract_feature_table(project, feature_table)

    def _extract_feature_table(
        self, project: str, feature_table: FeatureTable
    ) -> Iterator[TableMetadata]:
        columns = []
        for index, entity_name in enumerate(feature_table.entities):
            entity = self._client.get_entity(entity_name, project=project)
            columns.append(
                ColumnMetadata(
                    entity.name, entity.description, entity.value_type, index
                )
            )

        for index, feature in enumerate(feature_table.features):
            columns.append(
                ColumnMetadata(
                    feature.name,
                    None,
                    feature.dtype.name,
                    len(feature_table.entities) + index,
                )
            )

        yield TableMetadata(
            "feast",
            self._feast_service,
            project,
            feature_table.name,
            None,
            columns,
        )

        if self._describe_feature_tables:
            created_at = datetime.utcfromtimestamp(
                feature_table.created_timestamp.seconds
            )
            description = f"* Created at **{created_at}**\n"

            if feature_table.labels:
                description += "* Labels:\n"
                for key, value in feature_table.labels.items():
                    description += f"    * {key}: **{value}**\n"

            yield TableMetadata(
                "feast",
                self._feast_service,
                project,
                feature_table.name,
                description,
                description_source="feature_table_details",
            )

            yield TableMetadata(
                "feast",
                self._feast_service,
                project,
                feature_table.name,
                f'```\n{yaml.dump(feature_table.to_dict()["spec"]["batchSource"])}```',
                description_source="batch_source",
            )

            if feature_table.stream_source:
                yield TableMetadata(
                    "feast",
                    self._feast_service,
                    project,
                    feature_table.name,
                    f'```\n{yaml.dump(feature_table.to_dict()["spec"]["streamSource"])}```',
                    description_source="stream_source",
                )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

from databuilder.rest_api.base_rest_api_query import BaseRestApiQuery


class QueryMerger:
    """
    To be used in rest_api_query

    e.g. Assuming
            query_merger = QueryMerger(query_to_merge=spaces_query, merge_key='dashboard_group_id'),
         where spaces_query yields a record like
            {
             'dashboard_group_id': 'ggg',
             'dashboard_group': 'dashboard group'
            },
         and RestApiQuery's inner_rest_api_query.execute() returns a record of
            {
             'dashboard_id': 'ddd',
             'dashboard_name': 'dashboard_name',
             'dashboard_group_id': 'ggg'
             },
         the final yield record_dict from RestApiQuery(query_merger=query_merger).execute() will be
            {
             'dashboard_id': 'ddd',
             'dashboard_name': 'dashboard_name',
             'dashboard_group_id': 'ggg',
             'dashboard_group': 'dashboard group'
            }
    """
    def __init__(self,
                 query_to_merge: BaseRestApiQuery,
                 merge_key: str,
                 ) -> None:
        self._query_to_merge = query_to_merge
        self._merge_key = merge_key
        self._computed_query_result: Dict[Any, Any] = dict()

    def merge_into(self, record_dict: dict) -> None:
        """
        Merge results of query_to_merge into record_dict. Update record_dict in place.

        :param record_dict: the record_dict to be updated in place
        :return:
        """
        # compute query results for easy lookup later to find the exact record to merge
        if not self._computed_query_result:
            self._computed_query_result = self._compute_query_result()

        value_of_merge_key = record_dict.get(self._merge_key)
        record_dict_to_merge = self._computed_query_result.get(value_of_merge_key)
        if not record_dict_to_merge:
            raise Exception(f'{self._merge_key} {value_of_merge_key} not found in query_to_merge results')
        # we don't want the query merger to overwrite values of existing keys in record_dict
        filterd_record_dict_to_merge = {
            key: record_dict_to_merge[key] for key in record_dict_to_merge if key not in record_dict
        }
        record_dict.update(filterd_record_dict_to_merge)

    def _compute_query_result(self) -> Dict[Any, Any]:
        """
        Transform the query result to a dictionary.

        Assuming merge_key is 'dashboard_id' and self._query_to_merge.execute() returns
            iter([{'dashboard_id': 'd1', 'dashboard_name': 'n1'}, {'dashboard_id': 'd2', 'dashboard_name': 'n2'}]),
        the returned dict of this method will look like
            {
             'd1': {'dashboard_id': 'd1', 'dashboard_name': 'n1'},
             'd2': {'dashboard_id': 'd2', 'dashboard_name': 'n2'},
            }

        :return: a dictionary
        """
        computed_query_results: Dict[Any, Any] = dict()
        iterator = self._query_to_merge.execute()

        while True:
            try:
                record = next(iterator)
            except StopIteration:
                return computed_query_results

            value_of_merge_key = record[self._merge_key]
            if value_of_merge_key in computed_query_results:
                raise Exception(
                    f'merge_key {self._merge_key} value {value_of_merge_key} is not unique across the query results'
                )

            computed_query_results[value_of_merge_key] = record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (  # noqa: F401
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree  # noqa: F401

from databuilder.extractor.base_postgres_metadata_extractor import BasePostgresMetadataExtractor


class PostgresMetadataExtractor(BasePostgresMetadataExtractor):
    """
    Extracts Postgres table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """

    def get_sql_statement(self, use_catalog_as_cluster_name: bool, where_clause_suffix: str) -> str:
        if use_catalog_as_cluster_name:
            cluster_source = "c.table_catalog"
        else:
            cluster_source = f"'{self._cluster}'"

        return """
        SELECT
          {cluster_source} as cluster, c.table_schema as schema, c.table_name as name, pgtd.description as description
          ,c.column_name as col_name, c.data_type as col_type
          , pgcd.description as col_description, ordinal_position as col_sort_order
        FROM INFORMATION_SCHEMA.COLUMNS c
        INNER JOIN
          pg_catalog.pg_statio_all_tables as st on c.table_schema=st.schemaname and c.table_name=st.relname
        LEFT JOIN
          pg_catalog.pg_description pgcd on pgcd.objoid=st.relid and pgcd.objsubid=c.ordinal_position
        LEFT JOIN
          pg_catalog.pg_description pgtd on pgtd.objoid=st.relid and pgtd.objsubid=0
        {where_clause_suffix}
        ORDER by cluster, schema, name, col_sort_order ;
        """.format(
            cluster_source=cluster_source,
            where_clause_suffix=where_clause_suffix,
        )

    def get_scope(self) -> str:
        return 'extractor.postgres_metadata'



class AtlasRelationshipTypes:
    group_dashboard = 'DashboardGroup__Dashboard'
    dashboard_owner = 'Dashboard_Users_Owner'
    dashboard_query = 'Dashboard__DashboardQuery'
    query_chart = 'DashboardQuery__DashboardChart'
    dashboard_execution = 'Dashboard__DashboardExecution'
    table_dashboard = 'Table__Dashboard'


class AtlasSerializedRelationshipFields:
    relation_type = 'relationshipType'
    entity_type_1 = 'entityType1'
    qualified_name_1 = 'entityQualifiedName1'
    entity_type_2 = 'entityType2'
    qualified_name_2 = 'entityQualifiedName2'


class AtlasSerializedEntityFields:
    operation = 'operation'
    relationships = 'relationships'
    type_name = 'typeName'
    attributes = 'attributes'
    relationships_separator = '|'
    relationships_kv_separator = '#'


class AtlasSerializedEntityOperation:
    CREATE = 'CREATE'
    UPDATE = 'UPDATE'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


from functools import lru_cache
from typing import (
    Any, Dict, Iterator, List, Union,
)

from sqlalchemy.engine.url import make_url

from databuilder.extractor.dashboard.apache_superset.apache_superset_extractor import ApacheSupersetBaseExtractor
from databuilder.models.dashboard.dashboard_table import DashboardTable
from databuilder.models.table_metadata import TableMetadata


class ApacheSupersetTableExtractor(ApacheSupersetBaseExtractor):
    def _get_extract_iter(self) -> Iterator[Union[DashboardTable, None]]:
        dashboards: Dict[str, set] = dict()

        ids = self._get_dataset_ids()

        data = [(self._get_dataset_details(i), self._get_dataset_related_objects(i)) for i in ids]

        for entry in data:
            dataset_details, dataset_objects = entry

            database_id = self.get_nested_field(dataset_details, 'result.database.id')

            if database_id:
                database_details = self._get_database_details(database_id)

                sql = self.get_nested_field(dataset_details, 'result.sql') or ''

                # if sql exists then table_name cannot be associated with physical table in db
                if not len(sql) > 0:
                    uri = self.get_nested_field(database_details, 'result.sqlalchemy_uri')
                    database_spec = make_url(uri)

                    db = self.driver_mapping.get(database_spec.drivername, database_spec.drivername)
                    schema = database_spec.database

                    cluster = self.cluster_mapping.get(database_id, self.cluster)
                    tbl = self.get_nested_field(dataset_details, 'result.table_name')

                    table_key = TableMetadata.TABLE_KEY_FORMAT.format(db=db,
                                                                      cluster=cluster,
                                                                      schema=schema,
                                                                      tbl=tbl)

                    for dashboard in dataset_objects.get('dashboards', dict()).get('result', []):
                        dashboard_id = str(dashboard.get('id'))

                        if not dashboards.get(dashboard_id):
                            dashboards[dashboard_id] = set()

                        dashboards[dashboard_id].add(table_key)
                else:
                    pass
            else:
                continue

        for dashboard_id, table_keys in dashboards.items():
            table_metadata: Dict[str, Any] = {'dashboard_id': dashboard_id, 'table_ids': table_keys}

            table_metadata.update(**self.common_params)

            result = DashboardTable(**table_metadata)

            yield result

    def _get_dataset_ids(self) -> List[str]:
        url = self.build_full_url('api/v1/dataset')

        data = self.execute_query(url)

        return data.get('ids', [])

    def _get_dataset_details(self, dataset_id: str) -> Dict[str, Any]:
        url = self.build_full_url(f'api/v1/dataset/{dataset_id}')

        data = self.execute_query(url)

        return data

    def _get_dataset_related_objects(self, dataset_id: str) -> Dict[str, Any]:
        url = self.build_full_url(f'api/v1/dataset/{dataset_id}/related_objects')

        data = self.execute_query(url)

        return data

    @lru_cache(maxsize=512)
    def _get_database_details(self, database_id: str) -> Dict[str, Any]:
        url = self.build_full_url(f'api/v1/database/{database_id}')

        data = self.execute_query(url)

        return data

    @property
    def driver_mapping(self) -> Dict[str, str]:
        return dict(self.conf.get(self.DRIVER_TO_DATABASE_MAPPING))

    @property
    def cluster_mapping(self) -> Dict[str, str]:
        return dict(self.conf.get(self.DATABASE_TO_CLUSTER_MAPPING))



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from datetime import datetime
from typing import Any, Dict

from pyhocon import ConfigFactory, ConfigTree

from databuilder.transformer.base_transformer import Transformer

TIMESTAMP_FORMAT = 'timestamp_format'
FIELD_NAME = 'field_name'

LOGGER = logging.getLogger(__name__)

DEFAULT_CONFIG = ConfigFactory.from_dict({TIMESTAMP_FORMAT: '%Y-%m-%dT%H:%M:%S.%fZ'})


class TimestampStringToEpoch(Transformer):
    """
    Transforms string timestamp into epoch
    """

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf.with_fallback(DEFAULT_CONFIG)
        self._timestamp_format = self._conf.get_string(TIMESTAMP_FORMAT)
        self._field_name = self._conf.get_string(FIELD_NAME)

    def transform(self, record: Dict[str, Any]) -> Dict[str, Any]:
        timestamp_str = record.get(self._field_name, '')

        if not timestamp_str:
            return record

        try:
            utc_dt = datetime.strptime(timestamp_str, self._timestamp_format)
        except ValueError:
            # if the timestamp_str doesn't match format, no conversion, return initial result
            record[self._field_name] = 0
            return record

        record[self._field_name] = int((utc_dt - datetime(1970, 1, 1)).total_seconds())
        return record

    def get_scope(self) -> str:
        return 'transformer.timestamp_str_to_epoch'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import importlib
from collections import defaultdict
from typing import Any, List

from pyhocon import ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.badge import Badge, BadgeMetadata
from databuilder.models.table_lineage import ColumnLineage, TableLineage
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


def split_badge_list(badges: str, separator: str) -> List[str]:
    """
    Splits a string of badges into a list, removing all empty badges.
    """
    if badges is None:
        return []

    return [badge for badge in badges.split(separator) if badge]


class CsvExtractor(Extractor):
    # Config keys
    FILE_LOCATION = 'file_location'

    """
    An Extractor that extracts records via CSV.
    """

    def init(self, conf: ConfigTree) -> None:
        """
        :param conf:
        """
        self.conf = conf
        self.file_location = conf.get_string(CsvExtractor.FILE_LOCATION)

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
        self._load_csv()

    def _load_csv(self) -> None:
        """
        Create an iterator to execute sql.
        """
        if not hasattr(self, 'results'):
            with open(self.file_location, 'r') as fin:
                self.results = [dict(i) for i in csv.DictReader(fin)]

        if hasattr(self, 'model_class'):
            results = [self.model_class(**result)
                       for result in self.results]
        else:
            results = self.results
        self.iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self.iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.csv'


class CsvTableBadgeExtractor(Extractor):
    # Config keys
    TABLE_FILE_LOCATION = 'table_file_location'
    BADGE_FILE_LOCATION = 'badge_file_location'
    BADGE_SEPARATOR = 'badge_separator'

    """
    An Extractor that combines Table and Badge CSVs.
    """
    def init(self, conf: ConfigTree) -> None:
        self.conf = conf
        self.table_file_location = conf.get_string(CsvTableBadgeExtractor.TABLE_FILE_LOCATION)
        self.badge_file_location = conf.get_string(CsvTableBadgeExtractor.BADGE_FILE_LOCATION)
        self.badge_separator = conf.get_string(CsvTableBadgeExtractor.BADGE_SEPARATOR, default=',')
        self._load_csv()

    def _get_key(self,
                 db: str,
                 cluster: str,
                 schema: str,
                 tbl: str
                 ) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(db=db,
                                                     cluster=cluster,
                                                     schema=schema,
                                                     tbl=tbl)

    def _load_csv(self) -> None:
        with open(self.badge_file_location, 'r') as fin:
            self.badges = [dict(i) for i in csv.DictReader(fin)]
        # print("BADGES: " + str(self.badges))

        parsed_badges = defaultdict(list)
        for badge_dict in self.badges:
            db = badge_dict['database']
            cluster = badge_dict['cluster']
            schema = badge_dict['schema']
            table_name = badge_dict['table_name']
            id = self._get_key(db, cluster, schema, table_name)
            split_badges = split_badge_list(badges=badge_dict['name'],
                                            separator=self.badge_separator)
            for badge_name in split_badges:
                badge = Badge(name=badge_name, category=badge_dict['category'])
                parsed_badges[id].append(badge)

        with open(self.table_file_location, 'r') as fin:
            tables = [dict(i) for i in csv.DictReader(fin)]

        results = []
        for table_dict in tables:
            db = table_dict['database']
            cluster = table_dict['cluster']
            schema = table_dict['schema']
            table_name = table_dict['name']
            id = self._get_key(db, cluster, schema, table_name)
            badges = parsed_badges[id]

            if badges is None:
                badges = []
            badge_metadata = BadgeMetadata(start_label=TableMetadata.TABLE_NODE_LABEL,
                                           start_key=id,
                                           badges=badges)
            results.append(badge_metadata)
        self._iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.csvtablebadge'


class CsvTableColumnExtractor(Extractor):
    # Config keys
    TABLE_FILE_LOCATION = 'table_file_location'
    COLUMN_FILE_LOCATION = 'column_file_location'
    BADGE_SEPARATOR = 'badge_separator'

    """
    An Extractor that combines Table and Column CSVs.
    """

    def init(self, conf: ConfigTree) -> None:
        """
        :param conf:
        """
        self.conf = conf
        self.table_file_location = conf.get_string(CsvTableColumnExtractor.TABLE_FILE_LOCATION)
        self.column_file_location = conf.get_string(CsvTableColumnExtractor.COLUMN_FILE_LOCATION)
        self.badge_separator = conf.get_string(CsvTableBadgeExtractor.BADGE_SEPARATOR, default=',')
        self._load_csv()

    def _get_key(self,
                 db: str,
                 cluster: str,
                 schema: str,
                 tbl: str
                 ) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(db=db,
                                                     cluster=cluster,
                                                     schema=schema,
                                                     tbl=tbl)

    def _load_csv(self) -> None:
        """
        Create an iterator to execute sql.
        """
        with open(self.column_file_location, 'r') as fin:
            self.columns = [dict(i) for i in csv.DictReader(fin)]

        parsed_columns = defaultdict(list)
        for column_dict in self.columns:
            db = column_dict['database']
            cluster = column_dict['cluster']
            schema = column_dict['schema']
            table_name = column_dict['table_name']
            id = self._get_key(db, cluster, schema, table_name)
            split_badges = split_badge_list(badges=column_dict['badges'],
                                            separator=self.badge_separator)
            column = ColumnMetadata(
                name=column_dict['name'],
                description=column_dict['description'],
                col_type=column_dict['col_type'],
                sort_order=int(column_dict['sort_order']),
                badges=split_badges
            )
            parsed_columns[id].append(column)

        # Create Table Dictionary
        with open(self.table_file_location, 'r') as fin:
            tables = [dict(i) for i in csv.DictReader(fin)]

        results = []
        for table_dict in tables:
            db = table_dict['database']
            cluster = table_dict['cluster']
            schema = table_dict['schema']
            table_name = table_dict['name']
            id = self._get_key(db, cluster, schema, table_name)
            columns = parsed_columns[id]
            if columns is None:
                columns = []
            table = TableMetadata(database=table_dict['database'],
                                  cluster=table_dict['cluster'],
                                  schema=table_dict['schema'],
                                  name=table_dict['name'],
                                  description=table_dict['description'],
                                  columns=columns,
                                  # TODO: this possibly should parse stringified booleans;
                                  # right now it only will be false for empty strings
                                  is_view=bool(table_dict['is_view']),
                                  tags=table_dict['tags']
                                  )
            results.append(table)
        self._iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.csvtablecolumn'


class CsvTableLineageExtractor(Extractor):
    # Config keys
    TABLE_LINEAGE_FILE_LOCATION = 'table_lineage_file_location'

    """
    An Extractor that creates Table Lineage between two tables
    """

    def init(self, conf: ConfigTree) -> None:
        """
        :param conf:
        """
        self.conf = conf
        self.table_lineage_file_location = conf.get_string(CsvTableLineageExtractor.TABLE_LINEAGE_FILE_LOCATION)
        self._load_csv()

    def _load_csv(self) -> None:
        """
        Create an iterator to execute sql.
        """

        with open(self.table_lineage_file_location, 'r') as fin:
            self.table_lineage = [dict(i) for i in csv.DictReader(fin)]

        results = []
        for lineage_dict in self.table_lineage:
            source_table_key = lineage_dict['source_table_key']
            target_table_key = lineage_dict['target_table_key']
            lineage = TableLineage(
                table_key=source_table_key,
                downstream_deps=[target_table_key]
            )
            results.append(lineage)

        self._iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.csvtablelineage'


class CsvColumnLineageExtractor(Extractor):
    # Config keys
    COLUMN_LINEAGE_FILE_LOCATION = 'column_lineage_file_location'

    """
    An Extractor that creates Column Lineage between two columns
    """

    def init(self, conf: ConfigTree) -> None:
        """
        :param conf:
        """
        self.conf = conf
        self.column_lineage_file_location = conf.get_string(CsvColumnLineageExtractor.COLUMN_LINEAGE_FILE_LOCATION)
        self._load_csv()

    def _load_csv(self) -> None:
        """
        Create an iterator to execute sql.
        """

        with open(self.column_lineage_file_location, 'r') as fin:
            self.column_lineage = [dict(i) for i in csv.DictReader(fin)]

        results = []
        for lineage_dict in self.column_lineage:
            source_column_key = lineage_dict['source_column_key']
            target_column_key = lineage_dict['target_column_key']
            lineage = ColumnLineage(
                column_key=source_column_key,
                downstream_deps=[target_column_key]
            )
            results.append(lineage)

        self._iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.csvcolumnlineage'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Dict, Iterator,
)

from pyhocon import ConfigFactory, ConfigTree

import databuilder.extractor.dashboard.tableau.tableau_dashboard_constants as const
from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.tableau.tableau_dashboard_utils import (
    TableauDashboardUtils, TableauGraphQLApiExtractor,
)
from databuilder.extractor.restapi.rest_api_extractor import STATIC_RECORD_DICT
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel

LOGGER = logging.getLogger(__name__)


class TableauGraphQLApiQueryExtractor(TableauGraphQLApiExtractor):
    """
    Implements the extraction-time logic for parsing the GraphQL result and transforming into a dict
    that fills the DashboardQuery model. Allows workbooks to be exlcuded based on their project.
    """

    CLUSTER = const.CLUSTER
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS

    def execute(self) -> Iterator[Dict[str, Any]]:
        response = self.execute_query()

        for query in response['customSQLTables']:
            for workbook in query['downstreamWorkbooks']:
                if workbook['projectName'] not in \
                        self._conf.get_list(TableauGraphQLApiQueryExtractor.EXCLUDED_PROJECTS, []):
                    data = {
                        'dashboard_group_id': workbook['projectName'],
                        'dashboard_id': TableauDashboardUtils.sanitize_workbook_name(workbook['name']),
                        'query_name': query['name'],
                        'query_id': query['id'],
                        'query_text': query['query'],
                        'cluster': self._conf.get_string(TableauGraphQLApiQueryExtractor.CLUSTER)
                    }
                    yield data


class TableauDashboardQueryExtractor(Extractor):
    """
    Extracts metadata about the queries associated with Tableau workbooks.
    In terms of Tableau's Metadata API, these queries are called "custom SQL tables".
    However, not every workbook uses custom SQL queries, and most are built with a mixture of using the
    datasource fields directly and various "calculated" columns.
    This extractor iterates through one query at a time, yielding a new relationship for every downstream
    workbook that uses the query.
    """

    API_BASE_URL = const.API_BASE_URL
    API_VERSION = const.API_VERSION
    CLUSTER = const.CLUSTER
    EXCLUDED_PROJECTS = const.EXCLUDED_PROJECTS
    SITE_NAME = const.SITE_NAME
    TABLEAU_ACCESS_TOKEN_NAME = const.TABLEAU_ACCESS_TOKEN_NAME
    TABLEAU_ACCESS_TOKEN_SECRET = const.TABLEAU_ACCESS_TOKEN_SECRET
    VERIFY_REQUEST = const.VERIFY_REQUEST

    def init(self, conf: ConfigTree) -> None:
        self._conf = conf
        self.query = """query {
          customSQLTables {
            id
            name
            query
            downstreamWorkbooks {
              name
              projectName
            }
          }
        }"""

        self._extractor = self._build_extractor()

        transformers = []
        dict_to_model_transformer = DictToModel()
        dict_to_model_transformer.init(
            conf=Scoped.get_scoped_conf(self._conf, dict_to_model_transformer.get_scope()).with_fallback(
                ConfigFactory.from_dict(
                    {MODEL_CLASS: 'databuilder.models.dashboard.dashboard_query.DashboardQuery'})))
        transformers.append(dict_to_model_transformer)
        self._transformer = ChainedTransformer(transformers=transformers)

    def extract(self) -> Any:
        record = self._extractor.extract()
        if not record:
            return None

        return next(self._transformer.transform(record=record), None)

    def get_scope(self) -> str:
        return 'extractor.tableau_dashboard_query'

    def _build_extractor(self) -> TableauGraphQLApiQueryExtractor:
        """
        Builds a TableauGraphQLApiQueryExtractor. All data required can be retrieved with a single GraphQL call.
        :return: A TableauGraphQLApiQueryExtractor that provides dashboard query metadata.
        """
        extractor = TableauGraphQLApiQueryExtractor()
        tableau_extractor_conf = \
            Scoped.get_scoped_conf(self._conf, extractor.get_scope())\
                  .with_fallback(self._conf)\
                  .with_fallback(ConfigFactory.from_dict({TableauGraphQLApiExtractor.QUERY: self.query,
                                                          STATIC_RECORD_DICT: {'product': 'tableau'}
                                                          }
                                                         )
                                 )
        extractor.init(conf=tableau_extractor_conf)
        return extractor



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc

from pyhocon import ConfigTree

from databuilder import Scoped
from databuilder.utils.closer import Closer


class Job(Scoped):
    closer = Closer()

    """
    A Databuilder job that represents single work unit.
    """
    @abc.abstractmethod
    def init(self, conf: ConfigTree) -> None:
        pass

    @abc.abstractmethod
    def launch(self) -> None:
        """
        Launch a job
        :return: None
        """
        pass

    def get_scope(self) -> str:
        return 'job'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from databuilder.models.elasticsearch_document import ElasticsearchDocument


class UserESDocument(ElasticsearchDocument):
    """
    Schema for the Search index document for user
    """

    def __init__(self,
                 email: str,
                 first_name: str,
                 last_name: str,
                 full_name: str,
                 github_username: str,
                 team_name: str,
                 employee_type: str,
                 manager_email: str,
                 slack_id: str,
                 role_name: str,
                 is_active: bool,
                 total_read: int,
                 total_own: int,
                 total_follow: int,
                 ) -> None:
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.full_name = full_name
        self.github_username = github_username
        self.team_name = team_name
        self.employee_type = employee_type
        self.manager_email = manager_email
        self.slack_id = slack_id
        self.role_name = role_name
        self.is_active = is_active
        self.total_read = total_read
        self.total_own = total_own
        self.total_follow = total_follow



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterable, Iterator, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.table import TableUsage as RDSTableUsage

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.models.user import User


class ColumnReader(object):
    """
    A class represent user's read action on column. Implicitly assumes that read count is one.
    """

    def __init__(self,
                 database: str,
                 cluster: str,
                 schema: str,
                 table: str,
                 column: str,
                 user_email: str,
                 read_count: int = 1
                 ) -> None:
        self.database = database
        self.cluster = cluster
        self.schema = schema
        self.table = table
        self.column = column
        self.user_email = user_email
        self.read_count = int(read_count)

    def __repr__(self) -> str:
        return f"ColumnReader(database={self.database!r}, cluster={self.cluster!r}, " \
               f"schema={self.schema!r}, table={self.table!r}, column={self.column!r}, " \
               f"user_email={self.user_email!r}, read_count={self.read_count!r})"


class TableColumnUsage(GraphSerializable, TableSerializable):
    """
    A model represents user <--> column graph model
    Currently it only support to serialize to table level
    """
    TABLE_NODE_LABEL = TableMetadata.TABLE_NODE_LABEL
    TABLE_NODE_KEY_FORMAT = TableMetadata.TABLE_KEY_FORMAT

    USER_TABLE_RELATION_TYPE = 'READ'
    TABLE_USER_RELATION_TYPE = 'READ_BY'

    # Property key for relationship read, readby relationship
    READ_RELATION_COUNT = 'read_count'

    def __init__(self, col_readers: Iterable[ColumnReader]) -> None:
        for col_reader in col_readers:
            if col_reader.column != '*':
                raise NotImplementedError(f'Column is not supported yet {col_readers}')

        self.col_readers = col_readers
        self._node_iterator = self._create_node_iterator()
        self._rel_iter = self._create_rel_iterator()
        self._record_iter = self._create_record_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        for col_reader in self.col_readers:
            if col_reader.column == '*':
                # using yield for better memory efficiency
                user_node = User(email=col_reader.user_email).get_user_node()
                yield user_node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._rel_iter)
        except StopIteration:
            return None

    def _create_rel_iterator(self) -> Iterator[GraphRelationship]:
        for col_reader in self.col_readers:
            relationship = GraphRelationship(
                start_label=TableMetadata.TABLE_NODE_LABEL,
                start_key=self._get_table_key(col_reader),
                end_label=User.USER_NODE_LABEL,
                end_key=self._get_user_key(col_reader.user_email),
                type=TableColumnUsage.TABLE_USER_RELATION_TYPE,
                reverse_type=TableColumnUsage.USER_TABLE_RELATION_TYPE,
                attributes={
                    TableColumnUsage.READ_RELATION_COUNT: col_reader.read_count
                }
            )
            yield relationship

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        for col_reader in self.col_readers:
            if col_reader.column == '*':
                user_record = User(email=col_reader.user_email).get_user_record()
                yield user_record

            table_usage_record = RDSTableUsage(user_rk=self._get_user_key(col_reader.user_email),
                                               table_rk=self._get_table_key(col_reader),
                                               read_count=col_reader.read_count)
            yield table_usage_record

    def _get_table_key(self, col_reader: ColumnReader) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(db=col_reader.database,
                                                     cluster=col_reader.cluster,
                                                     schema=col_reader.schema,
                                                     tbl=col_reader.table)

    def _get_user_key(self, email: str) -> str:
        return User.get_user_model_key(email=email)

    def __repr__(self) -> str:
        return f'TableColumnUsage(col_readers={self.col_readers!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, List, Tuple, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.table import TableWatermark as RDSTableWatermark

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable


class Watermark(GraphSerializable, TableSerializable):
    """
    Table watermark result model.
    Each instance represents one row of table watermark result.
    """
    LABEL = 'Watermark'
    KEY_FORMAT = '{database}://{cluster}.{schema}' \
                 '/{table}/{part_type}/'
    WATERMARK_TABLE_RELATION_TYPE = 'BELONG_TO_TABLE'
    TABLE_WATERMARK_RELATION_TYPE = 'WATERMARK'

    def __init__(self,
                 create_time: str,
                 database: str,
                 schema: str,
                 table_name: str,
                 part_name: str,
                 part_type: str = 'high_watermark',
                 cluster: str = 'gold',
                 ) -> None:
        self.create_time = create_time
        self.database = database
        self.schema = schema
        self.table = table_name
        self.parts: List[Tuple[str, str]] = []

        if '=' not in part_name:
            raise Exception('Only partition table has high watermark')

        # currently we don't consider nested partitions
        idx = part_name.find('=')
        name, value = part_name[:idx], part_name[idx + 1:]
        self.parts = [(name, value)]
        self.part_type = part_type
        self.cluster = cluster
        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_next_record()

    def create_next_node(self) -> Union[GraphNode, None]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_watermark_model_key(self) -> str:
        return Watermark.KEY_FORMAT.format(database=self.database,
                                           cluster=self.cluster,
                                           schema=self.schema,
                                           table=self.table,
                                           part_type=self.part_type)

    def get_metadata_model_key(self) -> str:
        return f'{self.database}://{self.cluster}.{self.schema}/{self.table}'

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create watermark nodes
        :return:
        """
        for part in self.parts:
            part_node = GraphNode(
                key=self.get_watermark_model_key(),
                label=Watermark.LABEL,
                attributes={
                    'partition_key': part[0],
                    'partition_value': part[1],
                    'create_time': self.create_time
                }
            )
            yield part_node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relation map between watermark record with original table
        :return:
        """
        relation = GraphRelationship(
            start_key=self.get_watermark_model_key(),
            start_label=Watermark.LABEL,
            end_key=self.get_metadata_model_key(),
            end_label='Table',
            type=Watermark.WATERMARK_TABLE_RELATION_TYPE,
            reverse_type=Watermark.TABLE_WATERMARK_RELATION_TYPE,
            attributes={}
        )
        yield relation

    def _create_next_record(self) -> Iterator[RDSModel]:
        """
        Create watermark records
        """
        for part in self.parts:
            part_record = RDSTableWatermark(
                rk=self.get_watermark_model_key(),
                partition_key=part[0],
                partition_value=part[1],
                create_time=self.create_time,
                table_rk=self.get_metadata_model_key()
            )
            yield part_record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardOwner as RDSDashboardOwner

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.owner import Owner
from databuilder.models.user import User


class DashboardOwner(Owner):
    """
    A model that encapsulate Dashboard's owner.
    Note that it does not create new user as it has insufficient information about user but it builds relation
    between User and Dashboard
    """

    def __init__(self,
                 dashboard_group_id: str,
                 dashboard_id: str,
                 email: str,
                 product: Optional[str] = '',
                 cluster: str = 'gold',
                 **kwargs: Any
                 ) -> None:

        Owner.__init__(
            self,
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=product,
                cluster=cluster,
                dashboard_group=dashboard_group_id,
                dashboard_name=dashboard_id
            ),
            owner_emails=[email]
        )
        self._email = email
        self._record_iterator = self._create_record_iterator()

    # override this because we do not want to create new User nodes from this model
    def create_next_node(self) -> Union[GraphNode, None]:
        return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    # override this because we do not want to create new User rows from this model
    def _create_record_iterator(self) -> Iterator[RDSModel]:
        yield RDSDashboardOwner(
            user_rk=User.get_user_model_key(email=self._email),
            dashboard_rk=self.start_key,
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
import logging
from typing import (
    Any, Dict, Iterable, Iterator,
)

LOGGER = logging.getLogger(__name__)


class BaseRestApiQuery(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def execute(self) -> Iterator[Dict[str, Any]]:
        """
        Provides iterator of the records. It uses iterator so that it can stream the result.
        :return:
        """

        return iter([dict()])


class RestApiQuerySeed(BaseRestApiQuery):
    """
    A seed RestApiQuery.

    RestApiQuery is using decorator pattern where it needs to have a seed to begin with. RestApiQuerySeed is for
    RestApiQuery to start with.

    Example: see ModeDashboardExtractor._build_restapi_query
    """

    def __init__(self,
                 seed_record: Iterable[Dict[str, Any]]
                 ) -> None:
        self._seed_record = seed_record

    def execute(self) -> Iterator[Dict[str, Any]]:
        return iter(self._seed_record)


class EmptyRestApiQuerySeed(RestApiQuerySeed):
    """
    Sometimes there simply isn't a record to seed with.
    """

    def __init__(self) -> None:
        super(EmptyRestApiQuerySeed, self).__init__([{'empty_rest_api_query_seed': 1}])



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from typing import (
    Any, Iterator, List, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasDashboardTypes, AtlasTableKey
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardTable as RDSDashboardTable

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.utils.atlas import AtlasRelationshipTypes

LOGGER = logging.getLogger(__name__)


class DashboardTable(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    A model that link Dashboard with the tables used in various charts of the dashboard.
    Note that it does not create new dashboard, table as it has insufficient information but it builds relation
    between Tables and Dashboard
    """

    DASHBOARD_TABLE_RELATION_TYPE = 'DASHBOARD_WITH_TABLE'
    TABLE_DASHBOARD_RELATION_TYPE = 'TABLE_OF_DASHBOARD'

    def __init__(self,
                 dashboard_group_id: str,
                 dashboard_id: str,
                 table_ids: List[str],
                 product: Optional[str] = '',
                 cluster: str = 'gold',
                 **kwargs: Any
                 ) -> None:
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        # A list of tables uri used in the dashboard
        self._table_ids = table_ids
        self._product = product
        self._cluster = cluster

        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        if self._relation_iterator is None:
            return None
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        for table_id in self._table_ids:
            m = re.match(r'([^./]+)://([^./]+)\.([^./]+)\/([^./]+)', table_id)
            if m:
                relationship = GraphRelationship(
                    start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
                    end_label=TableMetadata.TABLE_NODE_LABEL,
                    start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                        product=self._product,
                        cluster=self._cluster,
                        dashboard_group=self._dashboard_group_id,
                        dashboard_name=self._dashboard_id
                    ),
                    end_key=TableMetadata.TABLE_KEY_FORMAT.format(
                        db=m.group(1),
                        cluster=m.group(2),
                        schema=m.group(3),
                        tbl=m.group(4)
                    ),
                    type=DashboardTable.DASHBOARD_TABLE_RELATION_TYPE,
                    reverse_type=DashboardTable.TABLE_DASHBOARD_RELATION_TYPE,
                    attributes={}
                )
                yield relationship

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        for table_id in self._table_ids:
            m = re.match(r'([^./]+)://([^./]+)\.([^./]+)\/([^./]+)', table_id)
            if m:
                yield RDSDashboardTable(
                    dashboard_rk=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                        product=self._product,
                        cluster=self._cluster,
                        dashboard_group=self._dashboard_group_id,
                        dashboard_name=self._dashboard_id
                    ),
                    table_rk=TableMetadata.TABLE_KEY_FORMAT.format(
                        db=m.group(1),
                        cluster=m.group(2),
                        schema=m.group(3),
                        tbl=m.group(4)
                    )
                )

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        pass

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        for table_id in self._table_ids:
            table_key = AtlasTableKey(table_id)
            table_relationship = AtlasRelationship(
                relationshipType=AtlasRelationshipTypes.table_dashboard,
                entityType1=table_key.get_details()['database'],
                entityQualifiedName1=table_key.qualified_name,
                entityType2=AtlasDashboardTypes.metadata,
                entityQualifiedName2=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                    product=self._product,
                    cluster=self._cluster,
                    dashboard_group=self._dashboard_group_id,
                    dashboard_name=self._dashboard_id
                ),
                attributes={}
            )
            yield table_relationship

    def __repr__(self) -> str:
        return f'DashboardTable({self._dashboard_group_id!r}, {self._dashboard_id!r}, ' \
               f'{self._product!r}, {self._cluster!r}, ({",".join(self._table_ids)!r}))'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
import os

from setuptools import find_packages, setup

__version__ = '4.5.3'

requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt')
with open(requirements_path) as requirements_file:
    requirements = requirements_file.readlines()

requirements_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements-dev.txt')
with open(requirements_path) as requirements_file:
    requirements_dev = requirements_file.readlines()

kafka = ['confluent-kafka==1.0.0']

cassandra = ['cassandra-driver==3.20.1']

glue = ['boto3==1.10.1']

snowflake = [
    'snowflake-connector-python',
    'snowflake-sqlalchemy'
]

athena = ['PyAthena[SQLAlchemy]>=1.0.0, <2.0.0']

# Python API client for google
# License: Apache Software License
# Upstream url: https://github.com/googleapis/google-api-python-client
bigquery = [
    'google-api-python-client>=1.6.0, <2.0.0dev',
    'google-auth-httplib2>=0.0.1',
    'google-auth>=1.0.0, <2.0.0dev'
]

jsonpath = ['jsonpath_rw==1.4.0']

db2 = [
    'ibm_db==3.0.1',
    'ibm-db-sa-py3==0.3.1-1'
]

dremio = [
    'pyodbc==4.0.30'
]

druid = [
    'pydruid'
]

spark = [
    'pyspark == 3.0.1'
]

neptune = [
    'amundsen-gremlin>=0.0.9',
    'Flask==1.0.2',
    'gremlinpython==3.4.3',
    'requests-aws4auth==0.9',
    'typing-extensions==3.7.4',
    'overrides==2.5',
    'boto3==1.10.1'
]

feast = [
    'feast==0.8.0'
]

atlas = [
    'pyatlasclient==1.1.2',
    'apache-atlas>=0.0.11'
]

rds = [
    'sqlalchemy>=1.3.6,<1.4',
    'mysqlclient>=1.3.6,<3'
]

all_deps = requirements + requirements_dev + kafka + cassandra + glue + snowflake + athena + \
    bigquery + jsonpath + db2 + dremio + druid + spark + feast + neptune + rds + atlas

setup(
    name='amundsen-databuilder',
    version=__version__,
    description='Amundsen Data builder',
    url='https://www.github.com/amundsen-io/amundsen/tree/main/databuilder',
    maintainer='Amundsen TSC',
    maintainer_email='amundsen-tsc@lists.lfai.foundation',
    packages=find_packages(exclude=['tests*']),
    dependency_links=[],
    install_requires=requirements,
    python_requires='>=3.6',
    extras_require={
        'all': all_deps,
        'dev': requirements_dev,
        'kafka': kafka,  # To use with Kafka source extractor
        'cassandra': cassandra,
        'glue': glue,
        'snowflake': snowflake,
        'athena': athena,
        'bigquery': bigquery,
        'jsonpath': jsonpath,
        'db2': db2,
        'dremio': dremio,
        'druid': druid,
        'neptune': neptune,
        'delta': spark,
        'feast': feast,
        'atlas': atlas,
        'rds': rds
    },
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
